with Ada.Text_IO;
with Ada.Strings.Fixed;
with Interfaces;
with Ada.Streams;
with Ada.Exceptions;
with System;
with Network_Utils;

with Data_Transport.Socket_Utilities;

package body Data_Transport.Socket_Client is

   procedure Handle_Data (Socket : GNAT.Sockets.Socket_Type;
                          Buffer_Set : Buffers.Buffer_Produce_Access;
                          Size : Interfaces.Unsigned_32);

   --------------------------
   -- Socket_Client_Task --
   --------------------------

   task body Socket_Client_Task is
      Address : GNAT.Sockets.Sock_Addr_Type;
      Socket : GNAT.Sockets.Socket_Type;
      use type Buffers.Buffer_Produce_Access;
   begin
      --  Ada.Text_IO.Put_Line ("socket receiver step 1");
      select
         accept Initialise (Host : String;
                            Port_Number : GNAT.Sockets.Port_Type) do
            --  Ada.Text_IO.Put_Line ("socket receiver step 1.1 " & Host);
            if Ada.Strings.Fixed.Index (Host, ".") /= 0 then
               Address.Addr := GNAT.Sockets.Inet_Addr (Host);
            else
               Address.Addr := GNAT.Sockets.Addresses
                 (GNAT.Sockets.Get_Host_By_Name (Host));
            end if;
            --  Ada.Text_IO.Put_Line ("socket receiver step 1.2");
            Address.Port := Port_Number;
            --  Ada.Text_IO.Put_Line ("socket receiver step 1.3");
            GNAT.Sockets.Create_Socket (Socket);
            --  Set_Socket_Option (Server, Option => (Name => Receive_Timeout,
            --  Timeout => 0.02));
            --  Ada.Text_IO.Put_Line ("socket receiver step 1.4");
         end Initialise;
      or
         terminate;
      end select;
      --  Ada.Text_IO.Put_Line ("socket receiver step 2");
      loop
         select
            accept Connect;
            --  Ada.Text_IO.Put_Line ("before connect");
            GNAT.Sockets.Connect_Socket (Socket, Address);
            --  Ada.Text_IO.Put_Line ("after connect");
            exit;
         or
            terminate;
         end select;
      end loop;
      --  Ada.Text_IO.Put_Line ("socket receiver step 3");
      loop
         declare
            use GNAT.Sockets;
            use type Interfaces.Unsigned_32;
         begin
            --  Ada.Text_IO.Put_Line ("socket receiver step 3.1");
            declare
               Buffer_Size : Interfaces.Unsigned_32;
               Data : Ada.Streams.Stream_Element_Array (1 .. 4);
               for Data'Address use Buffer_Size'Address;
            begin
               Socket_Utilities.Receive (Socket, Data);
               --  Ada.Text_IO.Put_Line
               --  ("data length : " &
               --       Network_Utils.From_Network (Buffer_Size)'Img);
               if Buffer_Size /= 0 then
                  Handle_Data (Socket, Buffer_Set, Buffer_Size);
               else
                  Ada.Text_IO.Put_Line ("Wait for Disconnect RdV");
                  select
                     accept Disconnect;
                     GNAT.Sockets.Close_Socket (Socket);
                     exit;
                  or
                     terminate;
                  end select;
               end if;
            end;
         exception
            when E : others =>
               Ada.Text_IO.Put_Line
                 ("task socket receiver exception : " &
                    Ada.Exceptions.Exception_Name (E) & "," &
                    Ada.Exceptions.Exception_Message (E));
               raise;
         end;
      end loop;
      --  Ada.Text_IO.Put_Line ("socket receiver step 4");
   exception
      when E : others =>
         Ada.Text_IO.Put_Line ("task socket receiver unattended exception : " &
                               Ada.Exceptions.Exception_Name (E) & "," &
                               Ada.Exceptions.Exception_Message (E));
   end Socket_Client_Task;

   procedure Handle_Data (Socket : GNAT.Sockets.Socket_Type;
                          Buffer_Set : Buffers.Buffer_Produce_Access;
                          Size : Interfaces.Unsigned_32) is
      Local_Size : constant Ada.Streams.Stream_Element_Offset :=
        Ada.Streams.Stream_Element_Offset (Network_Utils.From_Network (Size));
      subtype Data_Array is Ada.Streams.Stream_Element_Array (1 .. Local_Size);
   begin
      --  Ada.Text_IO.Put_Line ("buffer size" & Local_Size'Img);
      declare
         Buffer_Handle : Buffers.Buffer_Handle_Type;
         Loop_Exit : Boolean := False;
      begin
         loop
            exit when Loop_Exit;
            select
               Buffer_Set.Get_Free_Buffer (Buffer_Handle);
               declare
                  Buffer_Address : constant System.Address :=
                    Buffers.Get_Address (Buffer_Handle);
                  Data : Data_Array;
                  for Data'Address use Buffer_Address;
                  use GNAT.Sockets;
               begin
                  Socket_Utilities.Receive (Socket, Data);
               exception
                  when E : others =>
                     Ada.Text_IO.Put_Line
                       ("receive unattended exception : " &
                          Ada.Exceptions.Exception_Name (E) & "," &
                          Ada.Exceptions.Exception_Message (E));
                     raise;
               end;
               Buffers.Set_Used_Bytes (Buffer_Handle,
                                       Buffers.Buffer_Size_Type (Local_Size));
               Buffer_Set.Release_Free_Buffer (Buffer_Handle);
               Loop_Exit := True;
            or
               delay 0.1;
               --  Ada.Text_IO.Put_Line ("no block free");
               Loop_Exit := False;
            end select;
         end loop;
      end;
   exception
      when E : others =>
         Ada.Text_IO.Put_Line
           ("handle data unattended exception : " &
              Ada.Exceptions.Exception_Name (E) & "," &
              Ada.Exceptions.Exception_Message (E));
         raise;
   end Handle_Data;

end Data_Transport.Socket_Client;
