with Ada.Strings.Unbounded;
with Ada.Text_IO;

with GNAT.Sockets;

with Bufferized_Data_Transport;
--  with Broadcast_Buffers;
with Broadcast_TLs;
with Broadcast_Stuff;

with Buffers.Shared.Consume;
with Buffers.Shared.Produce;

with Data_Transport.Socket_Client;
with Data_Transport.Socket_Server;

with Global_Options;

package body DCOD.Ctl.Services is

   use Bufferized_Data_Transport;
   use Ada.Strings.Unbounded;

   Internal_Service_Error : exception;

   --------------------
   -- Request_Server --
   --------------------

   procedure Callback (Ressource : in out Broadcast_Stuff.Transport_Layer_Ressource_Type;
                       Address : GNAT.Sockets.Sock_Addr_Type);

   procedure Callback (Ressource : in out Broadcast_Stuff.Transport_Layer_Ressource_Type;
                       Address : GNAT.Sockets.Sock_Addr_Type) is
      Ip_String : constant String := GNAT.Sockets.Image (Address.Addr);
      Consumer : constant Buffers.Shared.Consume.Consume_Couple_Access :=
        new Buffers.Shared.Consume.Consume_Couple_Type;
      Buffer : Buffers.Shared.Consume.Consume_Type renames Consumer.Consumer;
      Server_Socket : constant Data_Transport.Socket_Server.Socket_Server_Access :=
        new Data_Transport.Socket_Server.Socket_Server_Task (Buffer'Access);
      Ressource_Name : constant String := Ressource.Get_Name;
      Qualified_Name : constant String := Ressource_Name & "@" & Ip_String;
      Buffer_Name : constant String := Ressource_Name (Ressource_Name'First + 5 .. Ressource_Name'Last);
   begin
      Ada.Text_IO.Put_Line ("callback for " & Ressource_Name & "," &
                              Buffer_Name & " " & Qualified_Name);
      Buffer.Set_Name (Buffer_Name);
      Consumer.Message_Handling.Start (1.0);
      Server_Socket.Initialise ("eth0", 8042);
      Server_Socket.Connect;
      Ressource.Qualified_Name := To_Unbounded_String (Qualified_Name);
      Transport_Layer_Shell.Request_Server_Second_Stage (Ressource_Name, Qualified_Name, Consumer, Server_Socket);
   end Callback;

   procedure Request_Server
     (Name : String;
      Buffer_Name : String)
   is
   begin
      Ada.Text_IO.Put_Line ("ctl declare output : enter");
      Broadcast_TLs.Ressource_Shell.Append ((Kind => Broadcast_Stuff.Transport_Layer_TCP,
                                             Name => To_Unbounded_String (Name),
                                             Host_Name => To_Unbounded_String (GNAT.Sockets.Host_Name),
                                             Process_Name => To_Unbounded_String (Global_Options.Get_Name),
                                             Error => False,
                                             Port => 8042,
                                             Qualified_Name => Null_Unbounded_String));
      Transport_Layer_Shell.Request_Server_First_Stage (Name, Buffer_Name);
      Ada.Text_IO.Put_Line ("ctl declare output : exit");
   end Request_Server;

   -------------------
   -- Request_Client --
   -------------------

   procedure Request_Client
     (Name : String;
      Buffer_Name : String;
      Server_Name : String)
   is
      Producer : constant Buffers.Shared.Produce.Produce_Couple_Access :=
        new Buffers.Shared.Produce.Produce_Couple_Type;
      Buffer : Buffers.Shared.Produce.Produce_Type renames Producer.Producer;
      Client_Socket : constant Data_Transport.Socket_Client.Socket_Client_Access :=
        new Data_Transport.Socket_Client.Socket_Client_Task (Buffer'Access);
   begin
      Ada.Text_IO.Put_Line ("ctl declare input : enter");
      Buffer.Set_Name (Buffer_Name);
      Ada.Text_IO.Put_Line ("ctl declare input : step 1");
      Producer.Message_Handling.Start (1.0);
      Ada.Text_IO.Put_Line ("ctl declare input : step 2");
      Broadcast_TLs.Ressource_Shell.Ask_For (Server_Name, Broadcast_Stuff.Transport_Layer_TCP);
      Ada.Text_IO.Put_Line ("ctl declare input : step 3");
      select
         Broadcast_TLs.Ressource_Shell.Wait_For_Answer;
         Ada.Text_IO.Put_Line ("get transport layer" & Server_Name);
      or
         delay 0.1;
         Ada.Text_IO.Put_Line ("ressource too long to answer");
         raise Internal_Service_Error with "ressource " & Server_Name & " doesn't seem to be available";
      end select;
      Ada.Text_IO.Put_Line ("ctl declare input : step 4");
      declare
         Ressource : constant Broadcast_Stuff.Transport_Layer_Ressource_Type :=
           Broadcast_TLs.Ressource_Shell.Get_Ressource (Server_Name);
      begin
         Ada.Text_IO.Put_Line ("TL ressource " & To_String (Ressource.Host_Name) &
                                 " " & To_String (Ressource.Qualified_Name) &
                                 Ressource.Port'Img);
         Client_Socket.Initialise (To_String (Ressource.Host_Name),
                                  GNAT.Sockets.Port_Type (Ressource.Port));
         Ada.Text_IO.Put_Line ("input socket : init done");
         Client_Socket.Connect;
         Ada.Text_IO.Put_Line ("input socket : connect");
         Transport_Layer_Shell.Request_Client (Name => Name,
                                              Buffer_Name => Buffer_Name,
                                              Server_Name => Server_Name,
                                              Host_Name => Ressource.Host_Name,
                                              Qualified_Name => Ressource.Qualified_Name,
                                              Producer => Producer,
                                              Client_Socket => Client_Socket);
      end;
   end Request_Client;

   -------------------
   -- Release_Server --
   -------------------

   procedure Release_Server (Name : String) is
      Consumer : Buffers.Shared.Consume.Consume_Couple_Access;
      Server_Socket : Data_Transport.Socket_Server.Socket_Server_Access;
      use type Buffers.Shared.Consume.Consume_Couple_Access;
      use type Data_Transport.Socket_Server.Socket_Server_Access;
   begin
      Ada.Text_IO.Put_Line (Name & " - delete_output ");
      Transport_Layer_Shell.Release_Server (Server_Name => Name,
                                           Consumer => Consumer,
                                           Server_Socket => Server_Socket);
      if Consumer = null or Server_Socket = null then
         Ada.Text_IO.Put_Line (Name & " - doesn't exist");
         return;
      end if;
      Server_Socket.Disconnect;
      Consumer.Message_Handling.Stop;
      Ada.Text_IO.Put_Line (Name & " consumer before delete");
      Consumer.Consumer.Delete;
      Ada.Text_IO.Put_Line (Name & " consumer after delete");
      Buffers.Shared.Consume.Free (Consumer);
      Data_Transport.Socket_Server.Free (Server_Socket);
   end Release_Server;

   ------------------
   -- Release_Client --
   ------------------

   procedure Release_Client (Name : String) is
      Producer : Buffers.Shared.Produce.Produce_Couple_Access;
      Client_Socket : Data_Transport.Socket_Client.Socket_Client_Access;
      Host_Name : Ada.Strings.Unbounded.Unbounded_String;
      Qualified_Name : Ada.Strings.Unbounded.Unbounded_String;
   begin
      Ada.Text_IO.Put_Line ("CTL delete input : " & Name);
      Transport_Layer_Shell.Release_Client (Name, Host_Name, Qualified_Name, Producer, Client_Socket);
      Ada.Text_IO.Put_Line ("CTL delete input step 1 " & To_String (Host_Name) &
                              " " & To_String (Qualified_Name));
      Client_Socket.Disconnect;
      Ada.Text_IO.Put_Line ("CTL delete input step 2");
      Producer.Message_Handling.Stop;
      Ada.Text_IO.Put_Line ("CTL delete input step 3");
      Producer.Producer.Delete;
      Ada.Text_IO.Put_Line ("CTL delete input step 4");
      Buffers.Shared.Produce.Free (Producer);
      Ada.Text_IO.Put_Line ("CTL delete input step 5");
      Data_Transport.Socket_Client.Free (Client_Socket);
      Ada.Text_IO.Put_Line ("CTL delete input exit");
   end Release_Client;

   ------------------
   -- Server_Exist --
   ------------------

   function Server_Exist (Name : String) return Boolean is
   begin
      return Transport_Layer_Shell.Server_Exist (Name);
   end Server_Exist;

   -----------------
   -- Client_Exist --
   -----------------

   function Client_Exist (Name : String) return Boolean is
   begin
      return Transport_Layer_Shell.Client_Exist (Name);
   end Client_Exist;

begin
   Broadcast_TLs.Ressource_Shell.Subscribe (Callback'Access);
end DCOD.Ctl.Services;
