pragma Warnings (Off);
with System.CRTL;
pragma Warnings (On);

package body Buffers.Local is

   -----------------------
   -- Local_Buffer_Type --
   -----------------------

   use Ada.Strings.Unbounded;

   protected body Local_Buffer_Type is

      procedure Release_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class);

      --------------
      -- Set_Name --
      --------------

      procedure Set_Name (Name : String) is
         Already_Initialised : exception;
      begin
         if Local_Buffer_Type.Name /= Null_Unbounded_String then
            raise Already_Initialised with "buffer already initialised to " &
              To_String (Local_Buffer_Type.Name);
         end if;
         Local_Buffer_Type.Name := To_Unbounded_String (Name);
      end Set_Name;

      function Get_Name return String is
      begin
         return To_String (Name);
      end Get_Name;

      function Compare (Name : String) return Boolean is
      begin
         return Name = Local_Buffer_Type.Name;
      end Compare;

      function Get_State return String is
      begin
         return "<buffer type=""local"" error=""" & Error_State'Img & """>" &
           "<name>" & To_String (Name) & "</name>" &
           "<size>" & Buffer_Size'Img & "</size>" &
           "<depth>" & Initial_Depth'Img & "</depth>" &
           "<free>" & Free_Buffers.Length'Img & "</free>" &
           "<full>" & Full_Buffers.Length'Img & "</full>" &
           "</buffer>";
      end Get_State;

      function No_Data return Boolean is
      begin
         return Initial_Depth = Natural (Free_Buffers.Length);
      end No_Data;

      function Free_Occupancy_Rate return Natural is
      begin
         return Natural (Free_Buffers.Length);
      end Free_Occupancy_Rate;

      function Free_Depth return Positive is
      begin
         return Initial_Depth;
      end Free_Depth;

      -----------
      -- Reset --
      -----------

      function Safe_Delete return Boolean is
      begin
         return No_Data;
      end Safe_Delete;

      procedure Delete is
      begin
         loop
            exit when Is_Empty (Full_Buffers);
            declare
               Buffer : Buffer_Description_Type;
            begin
               Buffer := First_Element (Full_Buffers);
               Delete_First (Full_Buffers);
               System.CRTL.free (Buffer.Address);
            end;
         end loop;
         loop
            exit when Is_Empty (Free_Buffers);
            declare
               Buffer : Buffer_Description_Type;
            begin
               Buffer := First_Element (Free_Buffers);
               Delete_First (Free_Buffers);
               System.CRTL.free (Buffer.Address);
            end;
         end loop;
         Clear (Full_Buffers);
         Clear (Free_Buffers);
         Initialised := False;
         Buffer_Size := 0;
      end Delete;

      procedure Reset is
         Temp_Buffer_Size : constant Buffer_Size_Type := Buffer_Size;
      begin
         if Safe_Delete then
            Delete;
         else
            raise Cant_Reset;
         end if;
         Initialise (Initial_Depth, Temp_Buffer_Size);
      end Reset;

      ----------------------
      -- Get_Buffers_Size --
      ----------------------

      function Get_Buffers_Size return Buffer_Size_Type is
      begin
         return Buffer_Size;
      end Get_Buffers_Size;

      ----------------
      -- Initialise --
      ----------------

      procedure Initialise
        (Buffer_Number : Positive;
         Size : Buffer_Size_Type)
      is
         use System.CRTL;
      begin
         if Initialised then
            return;
         end if;
         Clear (Full_Buffers);
         Clear (Free_Buffers);
         Initialised := True;
         Buffer_Size := Size;
         for I in 1 .. Buffer_Number loop
            declare
               Buffer : constant Buffer_Description_Type :=
                 (Address => malloc (size_t (Buffer_Size)),
                  Bytes => Buffer_Size,
                  ID => 0,
                  Nb_Ref => 0);
            begin
               Append (Free_Buffers, Buffer);
            end;
         end loop;
         Initial_Depth := Buffer_Number;
      end Initialise;

      ---------------------
      -- Get_Free_Buffer --
      ---------------------

      entry Get_Free_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      when not Is_Empty (Free_Buffers) is
         Buffer : Buffer_Description_Type;
      begin
         if Buffer_Handle.State /= Raw then
            raise Invalid_Buffer_State with
              Buffer_Handle.State'Img & " need a raw handle";
         end if;
         Buffer := First_Element (Free_Buffers);
         Delete_First (Free_Buffers);
         Buffer_Handle.State := In_Use_Filling;
         Buffer_Handle.Address := Buffer.Address;
         Buffer_Handle.Bytes_Available := Buffer.Bytes;
         Buffer_Handle.Bytes_Used := 0;
      end Get_Free_Buffer;

      entry Block_Free when not Is_Empty (Free_Buffers) is
      begin
         null;
      end Block_Free;

      -------------------------
      -- Release_Free_Buffer --
      -------------------------

      procedure Release_Free_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      is
      begin
         Release_Buffer (Buffer_Handle);
      end Release_Free_Buffer;

      -------------------
      -- Free_Is_Empty --
      -------------------

      function Free_Is_Empty return Boolean is
      begin
         return Is_Empty (Free_Buffers);
      end Free_Is_Empty;

      ---------------------
      -- Get_Full_Buffer --
      ---------------------

      entry Get_Full_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      when not Is_Empty (Full_Buffers) is
         Buffer : Buffer_Description_Type;
      begin
         if Buffer_Handle.State /= Raw then
            raise Invalid_Buffer_State with
              Buffer_Handle.State'Img & " need a raw handle";
         end if;
         Buffer := First_Element (Full_Buffers);
         Delete_First (Full_Buffers);
         Buffer_Handle.State := In_Use_Dumping;
         Buffer_Handle.Address := Buffer.Address;
         Buffer_Handle.Bytes_Available := 0;
         Buffer_Handle.Bytes_Used := Buffer.Bytes;
      end Get_Full_Buffer;

      entry Block_Full when not Is_Empty (Full_Buffers) is
      begin
         null;
      end Block_Full;

      -------------------------
      -- Release_Full_Buffer --
      -------------------------

      procedure Release_Full_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      is
      begin
         Release_Buffer (Buffer_Handle);
      end Release_Full_Buffer;

      -------------------
      -- Full_Is_Empty --
      -------------------

      function Full_Is_Empty return Boolean is
      begin
         return Is_Empty (Full_Buffers);
      end Full_Is_Empty;

      function Full_Size return Buffer_Size_Type is
      begin
         if Is_Empty (Full_Buffers) then
            return 0;
         end if;
         return First_Element (Full_Buffers).Bytes;
      end Full_Size;

      procedure Release_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class) is
         Buffer : Buffer_Description_Type;
         use type Common_Types.Buffer_Size_Type;
      begin
         case Buffer_Handle.State is
            when Raw | Clean =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Filling =>
               if Buffer_Handle.Bytes_Used = 0 then
                  raise Invalid_Call;
               else
                  Buffer := (Address => Buffer_Handle.Address,
                             Bytes => Buffer_Handle.Bytes_Used,
                             ID => 0,
                             Nb_Ref => 0);
                  Append (Full_Buffers, Buffer);
               end if;
            when In_Use_Dumping =>
               Buffer := (Address => Buffer_Handle.Address,
                          Bytes => Buffer_Size,
                          ID => 0,
                          Nb_Ref => 0);
               Append (Free_Buffers, Buffer);
         end case;
         Buffer_Handle.State := Clean;
      end Release_Buffer;

      procedure Release_Free_Buffer_Not_Used
        (Buffer_Handle : in out Buffer_Handle_Type'Class) is
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Dumping =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Filling =>
               Append (Free_Buffers,
                       (Address => Buffer_Handle.Address,
                        Bytes => Buffer_Size,
                        ID => 0,
                        Nb_Ref => 0));
         end case;
         Buffer_Handle.State := Clean;
      end Release_Free_Buffer_Not_Used;

      procedure Release_Full_Buffer_Not_Used
        (Buffer_Handle : in out Buffer_Handle_Type'Class) is
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Filling =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Dumping =>
               Prepend (Full_Buffers,
                        (Address => Buffer_Handle.Address,
                         Bytes => Buffer_Handle.Bytes_Used,
                         ID => Buffer_Handle.Buffer_ID,
                         Nb_Ref => 0));
         end case;
         Buffer_Handle.State := Clean;
      end Release_Full_Buffer_Not_Used;

      procedure Set_Policy (Policy : Common_Types.Policy_Type;
                            Value : Boolean) is
      begin
         if Available_Policies (Policy) then
            Policies (Policy) := Value;
         else
            raise Not_Supported_Policy with Policy'Img & " isn't supported";
         end if;
      end Set_Policy;

      procedure Set_Mode (Mode : Common_Types.Mode_Type;
                          Value : Boolean) is
      begin
         Modes (Mode) := Value;
      end Set_Mode;

      function Get_Policies return Policy_Array is
      begin
         return Policies;
      end Get_Policies;

      function Get_Available_Policies return Policy_Array is
      begin
         return Available_Policies;
      end Get_Available_Policies;

      function Is_In_Error_State return Boolean is
      begin
         return Error_State;
      end Is_In_Error_State;

      entry Is_Initialised when Initialised is
      begin
         null;
      end Is_Initialised;

   end Local_Buffer_Type;

end Buffers.Local;
