with Ada.Exceptions;
with Ada.Streams;
with Ada.Text_IO;

with GNAT.Sockets;

with Broadcast_Stuff;
with Passive_Directory;

with Dcod_Pmh_Service.Client;

with P42.Streams;

package body Buffers.Mailboxes is

   use Ada.Strings.Unbounded;

   ---------------
   -- Open_Send --
   ---------------

   procedure Open_Send
     (Mailbox : out Mailbox_Send_Type;
      Name : String;
      Pmh : String)
   is
   begin
      Mailbox.Name := To_Unbounded_String (Name);
      Mailbox.Associated_Pmh :=
        To_Unbounded_String (Broadcast_Stuff.Get_Process (Passive_Directory.Directory,
                                                          Broadcast_Stuff.Posix_Memory_Handler,
                                                          Pmh,
                                                          GNAT.Sockets.Host_Name));
      Dcod_Pmh_Service.Client.Request_Buffer (Name,
                                             Endpoint => To_String (Mailbox.Associated_Pmh));
      Mailbox.Production.Producer.Set_Name (Name);
      Mailbox.Production.Message_Handling.Start (1.0);
      select
         Mailbox.Production.Producer.Is_Initialised;
      or
         delay 3.0;
         raise Initialised_Too_Long with "open send too long to initialise";
      end select;
   end Open_Send;

   ------------------
   -- Open_Receive --
   ------------------

   procedure Open_Receive
     (Mailbox : out Mailbox_Receive_Type;
      Name : String;
      Pmh : String)
   is
   begin
      Mailbox.Name := To_Unbounded_String (Name);
      Mailbox.Associated_Pmh :=
        To_Unbounded_String (Broadcast_Stuff.Get_Process (Passive_Directory.Directory,
                                                          Broadcast_Stuff.Posix_Memory_Handler,
                                                          Pmh,
                                                          GNAT.Sockets.Host_Name));
      Dcod_Pmh_Service.Client.Provide_Buffer (Name,
                                              MAILBOX_SIZE,
                                              MAILBOX_DEPTH,
                                              Endpoint => To_String (Mailbox.Associated_Pmh));
      Mailbox.Consumption.Consumer.Set_Name (Name);
      Mailbox.Consumption.Message_Handling.Start (1.0);
   end Open_Receive;

   -----------
   -- Close --
   -----------

   procedure Close (Mailbox : in out Mailbox_Send_Type) is
   begin
      Ada.Text_IO.Put_Line ("close send mailbox");
      Mailbox.Production.Message_Handling.Stop;
      Ada.Text_IO.Put_Line ("close send mailbox step 1");
      Mailbox.Production.Producer.Delete;
      Ada.Text_IO.Put_Line ("close send mailbox step 2 " & To_String (Mailbox.Name));
      Ada.Text_IO.Put_Line ("close send mailbox step 3");
   end Close;

   procedure Close (Mailbox : in out Mailbox_Receive_Type) is
   begin
      Mailbox.Consumption.Message_Handling.Stop;
      Mailbox.Consumption.Consumer.Delete;
      Dcod_Pmh_Service.Client.Delete_Buffer (To_String (Mailbox.Name),
                                             To_String (Mailbox.Associated_Pmh));
   end Close;

   ------------------
   -- Send_Generic --
   ------------------

   procedure Send_Generic
     (Mailbox : in out Mailbox_Send_Type;
      Message : Message_Type)
   is
      Buffer_Handle : Buffers.Buffer_Handle_Type;
      use Buffers;
      Stream : aliased P42.Streams.Streamer_Type (MAILBOX_SIZE);
   begin
      Ada.Text_IO.Put_Line ("send generic 1" & Message'Size'Img);
      Stream.Initialise;
      Ada.Text_IO.Put_Line ("send generic 2");
      select
         Mailbox.Production.Producer.Get_Free_Buffer (Buffer_Handle);
         Ada.Text_IO.Put_Line ("send generic 3");
      or
         delay 1.0;
         Ada.Text_IO.Put_Line ("send generic 4");
         raise Get_Buffer_Too_Long with "get free buffer too long";
      end select;
      Ada.Text_IO.Put_Line ("send generic 5");
      Message_Type'Output (Stream'Access, Message);
      declare
         Data : Ada.Streams.Stream_Element_Array (Stream.First .. Stream.Last);
         for Data'Address use Get_Address (Buffer_Handle);
      begin
         Ada.Text_IO.Put_Line ("send generic 6");
         Data := Stream.Data (Data'Range);
         for I in Data'Range loop
            Ada.Text_IO.Put_Line (Data (I)'Img);
         end loop;
         Ada.Text_IO.Put_Line ("send generic 7" & Data'Length'Img);
         Buffer_Handle.Set_Used_Bytes (Buffer_Size_Type (Data'Length));
         Ada.Text_IO.Put_Line ("send generic 8");
      end;
      Mailbox.Production.Producer.Release_Free_Buffer (Buffer_Handle);
      Ada.Text_IO.Put_Line ("send generic 9");
   end Send_Generic;

   ---------------------
   -- Receive_Generic --
   ---------------------

   function Receive_Generic (Mailbox : in out Mailbox_Receive_Type) return Message_Type is
      Buffer_Handle : Buffers.Buffer_Handle_Type;
   begin
      Ada.Text_IO.Put_Line ("receive generic 1");
      select
         Mailbox.Consumption.Consumer.Get_Full_Buffer (Buffer_Handle);
         Ada.Text_IO.Put_Line ("receive generic 2");
      or
         delay 5.0;
         Ada.Text_IO.Put_Line ("receive generic 3");
         raise Get_Buffer_Too_Long with "get full buffer too long";
      end select;
      Ada.Text_IO.Put_Line ("receive generic 4");
      declare
         Data : Ada.Streams.Stream_Element_Array
           (1 .. Ada.Streams.Stream_Element_Offset (Buffer_Handle.Get_Used_Bytes));
         for Data'Address use Get_Address (Buffer_Handle);
         Stream : aliased P42.Streams.Streamer_Type (Data'Length);
      begin
         Ada.Text_IO.Put_Line ("receive generic 5");
         Ada.Text_IO.Put_Line ("receive generic 6");
         Stream.Data := Data;
         Stream.First := Data'First;
         Stream.Last := Data'Last;
         Ada.Text_IO.Put_Line ("receive generic 7" & Data'Length'Img);
         Mailbox.Consumption.Consumer.Release_Full_Buffer (Buffer_Handle);
         Ada.Text_IO.Put_Line ("receive generic 8");
         for I in Data'Range loop
            Ada.Text_IO.Put_Line (Data (I)'Img);
         end loop;
         declare
            Message : constant Message_Type := Message_Type'Input (Stream'Access);
         begin
            Ada.Text_IO.Put_Line ("receive generic 9" & Message'Size'Img);
            return Message;
         end;
      exception
         when E : others =>
            Ada.Text_IO.Put_Line ("exception name : " & Ada.Exceptions.Exception_Name (E));
            Ada.Text_IO.Put_Line ("exception message : " & Ada.Exceptions.Exception_Message (E));
            raise;
      end;

   end Receive_Generic;

end Buffers.Mailboxes;
