with System.Storage_Elements;

with POSIX.Permissions;
with POSIX.Shared_Memory_Objects;
with POSIX.Memory_Mapping;

with Uuid;

package body Buffers.Shared.Consume is

   ------------------
   -- Consume_Type --
   ------------------

   protected body Consume_Type is

      procedure Set_Host_Available (New_Value : Boolean) is
      begin
         Host_Available := New_Value;
      end Set_Host_Available;

      ---------------------
      -- Get_Full_Buffer --
      ---------------------

      entry Get_Full_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      when not Ask_Buffer_Done or
        not Is_Empty (Full_Buffers) or
        not Host_Available is
         Buffer : Buffer_Description_Type;
         use type Ada.Containers.Count_Type;
      begin
         if not Host_Available then
            raise Host_Not_Available with "get_full_buffer need an available host";
         end if;
         if Buffer_Handle.State /= Raw then
            raise Invalid_Buffer_State with
              Buffer_Handle.State'Img & " need a raw handle";
         end if;
         if Ask_Buffer_Done then
            Buffer := Full_Buffers.First_Element;
            Full_Buffers.Delete_First;
            Buffer_Handle.State := In_Use_Dumping;
            Buffer_Handle.Address := Buffer.Address;
            Buffer_Handle.Bytes_Available := 0;
            Buffer_Handle.Bytes_Used := Buffer.Bytes;
            Buffer_Handle.Buffer_ID := Buffer.ID;
            No_Buffer_In_Use := False;
            if Full_Buffers.Length = 0 then
               Ask_Buffer_Done := False;
            end if;
         else
            if On_Fly then
               Send (Server_Queue,
                     (Kind => Get_Spy,
                      Sender => Uuid_Image));
            else
               Send (Server_Queue,
                     (Kind => Get_Full,
                      Sender => Uuid_Image));
            end if;
            Ask_Buffer_Done := True;
            requeue Get_Full_Buffer with abort;
         end if;
      end Get_Full_Buffer;

      entry Block_Full when not Is_Empty (Full_Buffers) is
      begin
         null;
      end Block_Full;

      -------------------------
      -- Release_Full_Buffer --
      -------------------------

      procedure Release_Full_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      is
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Filling =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Dumping =>
               if On_Fly then
                  Send (Server_Queue,
                        (Kind => Release_Spy,
                         Bytes => 0,
                         ID => Buffer_Handle.Buffer_ID));
               else
                  Send (Server_Queue,
                        (Kind => Release_Full,
                         Bytes => 0,
                         ID => Buffer_Handle.Buffer_ID));
               end if;
         end case;
         Buffer_Handle.State := Clean;
         No_Buffer_In_Use := True;
      end Release_Full_Buffer;

      procedure Release_Full_Buffer_Not_Used
        (Buffer_Handle : in out Buffer_Handle_Type'Class) is
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Filling =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Dumping =>
               Prepend (Full_Buffers,
                        (Address => Buffer_Handle.Address,
                         Bytes => Buffer_Handle.Bytes_Used,
                         ID => Buffer_Handle.Buffer_ID,
                         Nb_Ref => 0));
         end case;
         Buffer_Handle.State := Clean;
         No_Buffer_In_Use := True;
      end Release_Full_Buffer_Not_Used;

      function Full_Is_Empty return Boolean is
      begin
         return Is_Empty (Full_Buffers);
      end Full_Is_Empty;

      function Full_Size return Buffer_Size_Type is
      begin
         if Is_Empty (Full_Buffers) then
            return 0;
         end if;
         return First_Element (Full_Buffers).Bytes;
      end Full_Size;

      --------------
      -- Set_Name --
      --------------

      procedure Set_Name (Name : String) is
         Uuid_Value : Uuid.Uuid_Type;
         use POSIX.Message_Queues;
         use Ada.Strings.Unbounded;
      begin
         if Consume_Type.Name /= Null_Unbounded_String then
            raise Already_Initialised with "name is alreadty set : " &
              To_String (Consume_Type.Name);
         end if;
         Consume_Type.Name := To_Unbounded_String (Name);
         Uuid.Uuid_Generate (Uuid_Value);
         Uuid_Image := Uuid.Image (Uuid_Value);

         Server_Queue := Open (POSIX.To_POSIX_String ("/" & Name),
                               POSIX.IO.Read_Write);
         Memory_Descriptor := POSIX.Shared_Memory_Objects.Open_Shared_Memory
           (POSIX.To_POSIX_String ("/" & Name),
            POSIX.IO.Read_Write);
         if On_Fly then
            Send (Server_Queue,
                  (Kind => Subscribe_Consume_On_Fly,
                   Sender => Uuid_Image));
         else
            Send (Server_Queue,
                  (Kind => Subscribe_Consume,
                   Sender => Uuid_Image));
         end if;
         Host_Available := True;
      end Set_Name;

      function Get_Name return String is
      begin
         return Ada.Strings.Unbounded.To_String (Name);
      end Get_Name;

      function Get_State return String is
         use Ada.Strings.Unbounded;
      begin
         return "<buffer type=""local"" error=""" & Error_State'Img & """>" &
           "<name>" & To_String (Name) & "</name>" &
           "<size>" & Buffer_Size'Img & "</size>" &
           "<depth>" & Initial_Depth'Img & "</depth>" &
           "<full>" & Full_Buffers.Length'Img & "</full>" &
           "</buffer>";
      end Get_State;

      function No_Data return Boolean is
      begin
         return Natural (Full_Buffers.Length) = 0 and No_Buffer_In_Use;
      end No_Data;

      function Compare (Name : String) return Boolean is
         use type Ada.Strings.Unbounded.Unbounded_String;
      begin
         return Name = Consume_Type.Name;
      end Compare;

      -----------
      -- Reset --
      -----------

      function Safe_Delete return Boolean is
      begin
         if not Full_Buffers.Is_Empty then
            return False;
         end if;
         return True;
      end Safe_Delete;

      procedure Delete is
      begin
         for Buffer of Buffers loop
            POSIX.Memory_Mapping.Unmap_Memory
              (Buffer.Address,
               System.Storage_Elements.Storage_Offset
                 (Buffer.Bytes));
         end loop;
         POSIX.IO.Close (Memory_Descriptor);
         Clear (Full_Buffers);
         Clear (Buffers);
         if On_Fly then
            POSIX.Message_Queues.Unlink_Message_Queue
              (POSIX.To_POSIX_String ("/consume_on_fly_" & Uuid_Image));
            Send (Server_Queue,
                  (Kind => Unsubscribe_Consume_On_Fly,
                   Sender => Uuid_Image));
         else
            POSIX.Message_Queues.Unlink_Message_Queue
              (POSIX.To_POSIX_String ("/consume_" & Uuid_Image));
            Send (Server_Queue,
                  (Kind => Unsubscribe_Consume,
                   Sender => Uuid_Image));
         end if;
      end Delete;

      procedure Reset is
      begin
         if not No_Buffer_In_Use then
            raise Cant_Reset;
         end if;
         if not Full_Buffers.Is_Empty then
            raise Cant_Reset;
         end if;
      end Reset;

      ----------------------
      -- Get_Buffers_Size --
      ----------------------

      function Get_Buffers_Size return Buffer_Size_Type is
      begin
         return Buffer_Size;
      end Get_Buffers_Size;

      procedure Append_Buffer (Buffer : Buffer_Description_Type) is
      begin
         Buffers.Append (Buffer);
         Initial_Depth := Initial_Depth + 1;
      end Append_Buffer;

      function Get_Uuid return Uuid_Image_Type is
      begin
         return Uuid_Image;
      end Get_Uuid;

      procedure Set_Size (Size : Buffer_Size_Type) is
      begin
         Buffer_Size := Size;
      end Set_Size;

      procedure Full_Available (Message : Message_Type) is
         Buffer : Buffer_Description_Type;
      begin
         Buffer := Find (Buffers, Message.ID);
         Buffer.Bytes := Message.Bytes;
         Append (Full_Buffers, Buffer);
      end Full_Available;

      function Get_Memory_Descriptor return POSIX.IO.File_Descriptor is
      begin
         return Memory_Descriptor;
      end Get_Memory_Descriptor;

      function Is_In_Error_State return Boolean is
      begin
         return Error_State;
      end Is_In_Error_State;

      entry Is_Initialised when Initialised is
      begin
         null;
      end Is_Initialised;

      procedure No_More_Buffers_Message is
      begin
         Initialised := True;
      end No_More_Buffers_Message;

      procedure Error_Message is
      begin
         Error_State := True;
      end Error_Message;

      procedure Set_On_Fly_Flag is
      begin
         On_Fly := True;
      end Set_On_Fly_Flag;

      function On_Fly_Set return Boolean is
      begin
         return On_Fly;
      end On_Fly_Set;

   end Consume_Type;

   task body Message_Handling_Task is
      Message_Check_Timeout : Duration;
      Consume_Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Attrs : POSIX.Message_Queues.Attributes;
      use POSIX.Message_Queues;
      Offset : POSIX.IO_Count := 0;
   begin
      select
         accept Start (Timeout : Duration) do
            Message_Check_Timeout := Timeout;
            Set_Max_Messages (Attrs, 64);
            Set_Message_Length (Attrs, Message_Size);
            if Wrapper.Consumer.On_Fly_Set then
               Consume_Queue := Open_Or_Create
                 (POSIX.To_POSIX_String ("/consume_on_fly_" &
                                           Wrapper.Consumer.Get_Uuid),
                  POSIX.IO.Read_Write,
                  POSIX.Permissions.Owner_Permission_Set,
                  Attrs => Attrs);
            else
               Consume_Queue := Open_Or_Create
                 (POSIX.To_POSIX_String ("/consume_" &
                                           Wrapper.Consumer.Get_Uuid),
                  POSIX.IO.Read_Write,
                  POSIX.Permissions.Owner_Permission_Set,
                  Attrs => Attrs);
            end if;
         end Start;
      or
         terminate;
      end select;
      loop
         select
            accept Stop;
            exit;
         else
            declare
               Address : System.Address;
               use type Common_Types.Buffer_Size_Type;
               use type POSIX.IO_Count;
               use type POSIX.Memory_Mapping.Protection_Options;
            begin
               declare
                  Message : constant Message_Type := Relative_Timed_Receive (Consume_Queue,
                                                                             Message_Check_Timeout);
               begin
                  case Message.Kind is
                     when Propagate_Delete =>
                        Wrapper.Consumer.Set_Host_Available (False);
                     when Full_Available =>
                        Wrapper.Consumer.Full_Available (Message);
                     when Spyed_Available =>
                        Wrapper.Consumer.Full_Available (Message);
                     when New_Buffer_Available =>
                        if Wrapper.Consumer.Get_Buffers_Size = 0 then
                           Wrapper.Consumer.Set_Size (Message.Bytes);
                        end if;
                        Address := POSIX.Memory_Mapping.Map_Memory
                          (System.Storage_Elements.Storage_Offset (Message.Bytes),
                           POSIX.Memory_Mapping.Allow_Read +
                             POSIX.Memory_Mapping.Allow_Write,
                           POSIX.Memory_Mapping.Map_Shared,
                           Wrapper.Consumer.Get_Memory_Descriptor,
                           Offset);
                        Offset := Offset + POSIX.IO_Count (Message.Bytes);
                        Wrapper.Consumer.Append_Buffer
                          ((Address => Address,
                            Bytes => Message.Bytes,
                            ID => Message.ID,
                            Nb_Ref => 0));
                     when No_More_Buffers =>
                        Wrapper.Consumer.No_More_Buffers_Message;
                     when Error =>
                        Wrapper.Consumer.Error_Message;
                     when Subscribe_Consume .. Release_Spy |
                       Free_Available |
                       Unsubscribe_Consume .. Unsubscribe_Consume_On_Fly =>
                        null;
                  end case;
               end;
            exception
               when POSIX.POSIX_Error =>
                  null;
            end;
         end select;
      end loop;
   end Message_Handling_Task;

   procedure Clean (Vector : in out Consumer_Vectors.Vector) is
   begin
      for Couple of Vector loop
         Free (Couple);
      end loop;
      Vector.Clear;
   end Clean;

end Buffers.Shared.Consume;
