with Ada.Exceptions;
with Ada.Text_IO;
with System.Storage_Elements;

with POSIX.Shared_Memory_Objects;
with POSIX.Permissions;
with POSIX.Memory_Mapping;

package body Buffers.Shared.Host is

   ---------------
   -- Host_Type --
   ---------------

   use Ada.Strings.Unbounded;

   protected body Host_Type is

      --------------
      -- Set_Name --
      --------------

      procedure Set_Name (Name : String) is
      begin
         if Host_Type.Name /= Null_Unbounded_String then
            raise Already_Initialised with "buffer already initialised to " &
              To_String (Host_Type.Name);
         end if;
         Host_Type.Name := To_Unbounded_String (Name);
      end Set_Name;

      function Get_Name return String is
      begin
         return To_String (Name);
      end Get_Name;

      function Get_State return String is
      begin
         return "<buffer type=""host"" nb_ref=""" & Nb_Ref'Img & """>" &
           "<name>" & To_String (Name) & "</name>" &
           "<size>" & Buffer_Size'Img & "</size>" &
           "<depth>" & Initial_Depth'Img & "</depth>" &
           "<free>" & Free_Buffers.Length'Img & "</free>" &
           "<full>" & Full_Buffers.Length'Img & "</full>" &
           "<spy>" & Spyed_Buffers.Length'Img & "</spy>" &
           "</buffer>";
      end Get_State;

      function No_Data return Boolean is
      begin
         return Initial_Depth = Natural (Free_Buffers.Length);
      end No_Data;

      function Compare (Name : String) return Boolean is
      begin
         return Name = Host_Type.Name;
      end Compare;

      -----------
      -- Reset --
      -----------

      function Safe_Delete return Boolean is
         use type Ada.Containers.Count_Type;
      begin
         return Nb_Ref = 0;
      end Safe_Delete;

      function Is_Propagation_Needed return Boolean is
      begin
         return Need_Propagation;
      end Is_Propagation_Needed;

      entry Delete when Nb_Ref = 0 or not Need_Propagation is
      begin
         if Nb_Ref /= 0 then
            Need_Propagation := True;
            requeue Delete with abort;
         end if;
         POSIX.IO.Close (Memory_Descriptor);
         POSIX.Shared_Memory_Objects.Unlink_Shared_Memory
           (POSIX.To_POSIX_String ("/" & To_String (Name)));
         Clear (Free_Buffers);
         Clear (Full_Buffers);
         POSIX.Message_Queues.Unlink_Message_Queue
           (POSIX.To_POSIX_String ("/" & To_String (Name)));
      end Delete;

      procedure Reset is
      begin
         if Nb_Ref /= 0 then
            raise Cant_Reset;
         end if;
         Full_Buffers.Clear;
         Free_Buffers.Clear;
         Spyed_Buffers.Clear;
         for Buffer of Ordered_Buffers loop
            Free_Buffers.Append (Buffer);
         end loop;
      end Reset;

      ----------------------
      -- Get_Buffers_Size --
      ----------------------

      function Get_Buffers_Size return Buffer_Size_Type is
      begin
         return Buffer_Size;
      end Get_Buffers_Size;

      ----------------
      -- Initialise --
      ----------------

      procedure Initialise
        (Buffer_Number : Positive;
         Size : Buffer_Size_Type)
      is
         Local_Full_Size : POSIX.IO_Count;
         POSIX_Size : constant POSIX.IO_Count := POSIX.IO_Count (Size);
         use type POSIX.Memory_Mapping.Protection_Options;
         use type POSIX.IO_Count;
         use type System.Storage_Elements.Storage_Offset;
         use type Common_Types.Buffer_Size_Type;
      begin
         if Size mod 4096 /= 0 then
            raise Not_Page_Aligned_Size with
              "shared buffer size should be aligned to a 4096 boundary";
         end if;
         if Name = Null_Unbounded_String then
            Name := To_Unbounded_String ("unnamed");
         end if;
         Local_Full_Size := POSIX.IO_Count (Buffer_Number) * POSIX_Size;
         Memory_Descriptor :=
           POSIX.Shared_Memory_Objects.Open_Or_Create_Shared_Memory
           (POSIX.To_POSIX_String ("/" & To_String (Name)),
            POSIX.IO.Read_Write,
            POSIX.Permissions.Owner_Permission_Set);
         POSIX.IO.Truncate_File (Memory_Descriptor, Local_Full_Size);
         for I in 1 .. Buffer_Number loop
            Append (Free_Buffers,
                    (Address => System.Null_Address,
                     Bytes => Size,
                     ID => Buffer_ID_Type (I),
                     Nb_Ref => 0));
            Append (Ordered_Buffers,
                    (Address => System.Null_Address,
                     Bytes => Size,
                     ID => Buffer_ID_Type (I),
                     Nb_Ref => 0));
         end loop;
         Buffer_Size := Size;
         Initial_Depth := Buffer_Number;
      end Initialise;

      procedure Set_Policy (Policy : Common_Types.Policy_Type;
                            Value : Boolean) is
      begin
         if Available_Policies (Policy) then
            Policies (Policy) := Value;
         else
            raise Not_Supported_Policy with Policy'Img & " isn't supported";
         end if;
      end Set_Policy;

      procedure Set_Mode (Mode : Common_Types.Mode_Type;
                          Value : Boolean) is
      begin
         Modes (Mode) := Value;
      end Set_Mode;

      function Get_Policies return Policy_Array is
      begin
         return Policies;
      end Get_Policies;

      function Get_Available_Policies return Policy_Array is
      begin
         return Available_Policies;
      end Get_Available_Policies;

      function Get_Message_Task return Message_Handling_Access is
      begin
         return Messages_Handling;
      end Get_Message_Task;

      procedure Send_Buffer_Description
        (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
         With_Full : Boolean) is
      begin
         for Buffer_Description of Ordered_Buffers loop
            declare
               Message_To_Return : constant Message_Type :=
                 (Kind => New_Buffer_Available,
                  ID => Buffer_Description.ID,
                  Bytes => Buffer_Description.Bytes);
            begin
               Send (Queue, Message_To_Return);
            end;
         end loop;
         Send (Queue,
               (Kind => No_More_Buffers));
         if With_Full then
            for Buffer_Description of Full_Buffers loop
               Buffer_Description.Nb_Ref := Buffer_Description.Nb_Ref + 1;
               Send (Queue,
                     (Kind => Full_Available,
                      ID => Buffer_Description.ID,
                      Bytes => Buffer_Description.Bytes));
            end loop;
         end if;
      end Send_Buffer_Description;

      procedure Add_Reference is
      begin
         Nb_Ref := Nb_Ref + 1;
      end Add_Reference;

      procedure Remove_Reference is
      begin
         Nb_Ref := Nb_Ref - 1;
      end Remove_Reference;

      procedure Release_Free
        (Message : Message_Type;
         Consumers : Queue_Vectors.Vector) is
         First_Consumer : Unique_Queue_Type;
         First_Producer : Unique_Queue_Type;
         use type Common_Types.Buffer_Size_Type;
         Buffer : Buffer_Description_Type;
      begin
         Buffer := Find (Ordered_Buffers, Message.ID);
         if Message.Bytes = 0 or Modes (Common_Types.No_Transmit) then
            --  Ada.Text_IO.Put_Line ("release free #1");
            Append (Free_Buffers, Buffer);
            if not Pending_Producers.Is_Empty then
               First_Producer := Pending_Producers.First_Element;
               Pending_Producers.Delete_First;
               Find_And_Remove (Free_Buffers, Message.ID, Buffer);
               --  Ada.Text_IO.Put_Line ("release free sending 1");
               Send (First_Producer.Queue,
                     (Free_Available,
                      Bytes => Buffer_Size,
                      ID => Buffer.ID));
            end if;
         else
            Buffer.Bytes := Message.Bytes;
            if Policies (Common_Types.Consume_Once) then
               Buffer.Nb_Ref := 1;
            else
               Buffer.Nb_Ref := Natural (Consumers.Length);
            end if;
            Append (Full_Buffers, Buffer);
            if Policies (Common_Types.Consume_Once) then
               if not Pending_Consumers.Is_Empty then
                  First_Consumer := Pending_Consumers.First_Element;
                  Pending_Consumers.Delete_First;
                  Send (First_Consumer.Queue,
                        (Kind => Full_Available,
                         ID => Message.ID,
                         Bytes => Message.Bytes));
               end if;
            else
               for Consumer of Consumers loop
                  Send (Consumer.Queue,
                        (Kind => Full_Available,
                         ID => Message.ID,
                         Bytes => Message.Bytes));
               end loop;
            end if;
         end if;
      end Release_Free;

      procedure Append_Pending_Consumer (Consumer : Unique_Queue_Type) is
         Buffer : Buffer_Description_Type;
         First_Consumer : Unique_Queue_Type;
      begin
         if not Policies (Common_Types.Consume_Once) then
            raise Incompatible_Call_With_Policy with
              "consume_once policy not enabled";
         end if;
         Pending_Consumers.Append (Consumer);
         if not Full_Buffers.Is_Empty then
            Buffer := Full_Buffers.First_Element;
            First_Consumer := Pending_Consumers.First_Element;
            Pending_Consumers.Delete_First;
            Send (First_Consumer.Queue,
                  (Kind => Full_Available,
                   ID => Buffer.ID,
                   Bytes => Buffer.Bytes));
         end if;
      end Append_Pending_Consumer;

      procedure Append_Pending_Consumer_On_Fly
        (Consumer : Unique_Queue_Type) is
         Buffer : Buffer_Description_Type;
         First_Consumer : Unique_Queue_Type;
      begin
         --  Ada.Text_IO.Put_Line ("append consumer on fly #0");
         if not Policies (Common_Types.Consume_On_Fly) then
            raise Incompatible_Call_With_Policy with
              "consume_on_fly policy not enabled";
         end if;
         --  Ada.Text_IO.Put_Line ("append consumer on fly #1");
         Pending_Consumers_On_Fly.Append (Consumer);
         --  Ada.Text_IO.Put_Line ("append consumer on fly #2");
         if not Spyed_Buffers.Is_Empty then
            --  Ada.Text_IO.Put_Line ("append consumer on fly #3");
            Buffer := Spyed_Buffers.First_Element;
            if Buffer.Nb_Ref = 0 then
               First_Consumer := Pending_Consumers_On_Fly.First_Element;
               Pending_Consumers_On_Fly.Delete_First;
               Send (First_Consumer.Queue,
                     (Kind => Spyed_Available,
                      ID => Buffer.ID,
                      Bytes => Buffer.Bytes));
               Spyed_Buffers.Delete_First;
               Buffer.Nb_Ref := 1;
               Spyed_Buffers.Append (Buffer);
               --  Ada.Text_IO.Put_Line ("append consumer on fly #4");
            end if;
         end if;
         --  Ada.Text_IO.Put_Line ("append consumer on fly #5");
      end Append_Pending_Consumer_On_Fly;

      procedure Append_Pending_Producer (Producer : Unique_Queue_Type) is
         Buffer : Buffer_Description_Type;
         First_Producer : Unique_Queue_Type;
      begin
         Pending_Producers.Append (Producer);
         if not Free_Buffers.Is_Empty then
            Buffer := Free_Buffers.First_Element;
            First_Producer := Pending_Producers.First_Element;
            Pending_Producers.Delete_First;
            Free_Buffers.Delete_First;
            --  Ada.Text_IO.Put_Line ("append sending");
            Send (First_Producer.Queue,
                  (Kind => Free_Available,
                   ID => Buffer.ID,
                   Bytes => Buffer_Size));
         end if;
      end Append_Pending_Producer;

      procedure Put_Back_In_Free_Queue
        (Buffer : in out Buffer_Description_Type) is
         First_Producer : Unique_Queue_Type;
      begin
         --  Ada.Text_IO.Put_Line ("put back in free queue #0" &
         --                          Free_Buffers.Length'Img);
         Buffer.Bytes := Buffer_Size;
         Free_Buffers.Append (Buffer);
         --  Ada.Text_IO.Put_Line ("put back in free queue #1");
         --  Ada.Text_IO.Put_Line ("put back in free queue 1 " &
         --                          Pending_Producers.Is_Empty'Img);
         if not Pending_Producers.Is_Empty then
            First_Producer := Pending_Producers.First_Element;
            Pending_Producers.Delete_First;
            Free_Buffers.Delete_Last;
            --  Ada.Text_IO.Put_Line ("put back in free queue sending 1");
            Send (First_Producer.Queue,
                  (Kind => Free_Available,
                   Bytes => Buffer_Size,
                   ID => Buffer.ID));
         end if;
         --  Ada.Text_IO.Put_Line ("put back in free queue 3-" &
         --                          Free_Buffers.Length'Img);
      end Put_Back_In_Free_Queue;

      procedure Release_Full (Message : Message_Type) is
         Buffer, First_Buffer : Buffer_Description_Type;
         First_Consumer : Unique_Queue_Type;
         Spyed_Buffer_Length : Natural :=
           Natural (Spyed_Buffers.Length);
      begin
         --  Ada.Text_IO.Put_Line ("release full enter");
         Find_And_Remove (Full_Buffers, Message.ID, Buffer);
         --  Ada.Text_IO.Put_Line ("release full step 1-" & Buffer.Nb_Ref'Img);
         Buffer.Nb_Ref := Buffer.Nb_Ref - 1;
         --  Ada.Text_IO.Put_Line ("release full step 2");
         if Buffer.Nb_Ref = 0 then
            if Spyed_Buffer_Length = Spyed_Limit and
              Policies (Common_Types.Consume_On_Fly) and
              Modes (Common_Types.On_Fly_Refresh)
            then
               First_Buffer := Spyed_Buffers.First_Element;
               Spyed_Buffers.Delete_First;
               Spyed_Buffer_Length := Spyed_Buffer_Length - 1;
               Put_Back_In_Free_Queue (First_Buffer);
            end if;
            if Policies (Common_Types.Consume_On_Fly) and
              Spyed_Buffer_Length < Spyed_Limit
            then
               Spyed_Buffers.Append (Buffer);
               if not Pending_Consumers_On_Fly.Is_Empty then
                  First_Consumer := Pending_Consumers_On_Fly.First_Element;
                  Pending_Consumers_On_Fly.Delete_First;
                  --  Ada.Text_IO.Put_Line ("release full sending spyed");
                  Send (First_Consumer.Queue,
                        (Kind => Spyed_Available,
                         Bytes => Buffer.Bytes,
                         ID => Buffer.ID));
                  Spyed_Buffers.Delete_Last;
                  Buffer.Nb_Ref := 1;
                  Spyed_Buffers.Append (Buffer);
               end if;
            else
               Put_Back_In_Free_Queue (Buffer);
            end if;
         else
            Full_Buffers.Append (Buffer);
         end if;
      end Release_Full;

      procedure Release_Spy (Message : Message_Type) is
         Buffer : Buffer_Description_Type;
         First_Producer : Unique_Queue_Type;
      begin
         --  Ada.Text_IO.Put_Line ("release spy #0");
         Find_And_Remove (Spyed_Buffers, Message.ID, Buffer);
         Buffer.Bytes := Buffer_Size;
         Free_Buffers.Append (Buffer);
         --  Ada.Text_IO.Put_Line ("release spy #1");
         --  Ada.Text_IO.Put_Line ("release spy step 1 " &
         --                          Pending_Producers.Is_Empty'Img);
         if not Pending_Producers.Is_Empty then
            First_Producer := Pending_Producers.First_Element;
            Pending_Producers.Delete_First;
            Free_Buffers.Delete_Last;
            --  Ada.Text_IO.Put_Line ("release spy sending 1");
            Send (First_Producer.Queue,
                  (Kind => Free_Available,
                   Bytes => Buffer_Size,
                   ID => Buffer.ID));
         end if;
      end Release_Spy;

      function Is_In_Error_State return Boolean is
      begin
         return False;
      end Is_In_Error_State;

      entry Is_Initialised when Initialised is
      begin
         null;
      end Is_Initialised;

      procedure Set_On_Fly_Limit (Limit : Positive) is
      begin
         Spyed_Limit := Limit;
      end Set_On_Fly_Limit;

   end Host_Type;

   task body Message_Handling_Task is
      Local_Host : Host_Access;
      Message_Check_Timeout : Duration;
      Server_Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Attrs : POSIX.Message_Queues.Attributes;
      Consumers : Queue_Vectors.Vector;
      Consumers_On_Fly : Queue_Vectors.Vector;
      Producers : Queue_Vectors.Vector;
      use type POSIX.Error_Code;
   begin
      select
         accept Start (Host_Item : Host_Access;
                       Timeout : Duration) do
            Local_Host := Host_Item;
            Message_Check_Timeout := Timeout;
            POSIX.Message_Queues.Set_Max_Messages (Attrs, 64);
            POSIX.Message_Queues.Set_Message_Length (Attrs, Message_Size);
            Server_Queue := POSIX.Message_Queues.Open_Or_Create
              (POSIX.To_POSIX_String ("/" & Host_Item.Get_Name),
               POSIX.IO.Read_Write,
               POSIX.Permissions.Owner_Permission_Set,
               POSIX.IO.Empty_Set,
               Attrs => Attrs);
         end Start;
      or
         terminate;
      end select;
      loop
         select
            accept Stop;
            exit;
         else
            loop
               begin
                  if Local_Host.Is_Propagation_Needed then
                     for Consumer of Consumers loop
                        Send (Consumer.Queue, (Kind => Propagate_Delete));
                     end loop;
                     for Producer of Producers loop
                        Send (Producer.Queue, (Kind => Propagate_Delete));
                     end loop;
                  end if;
                  declare
                     Message : constant Message_Type := Relative_Timed_Receive (Server_Queue,
                                                                                Message_Check_Timeout);
                  begin
                     --  Ada.Text_IO.Put_Line (Message.Kind'Img);
                     case Message.Kind is
                        when Subscribe_Consume =>
                           loop
                              declare
                                 use POSIX.Message_Queues;
                                 Consume_Queue : Message_Queue_Descriptor;
                              begin
                                 Consume_Queue := Open (POSIX.To_POSIX_String
                                                          ("/consume_" &
                                                             Message.Sender),
                                                        POSIX.IO.Read_Write);
                                 Consumers.Append ((Queue => Consume_Queue,
                                                    Uuid => Message.Sender));
                                 Local_Host.Add_Reference;
                                 if Local_Host.Get_Policies (Common_Types.Consume_Once) then
                                    Local_Host.Send_Buffer_Description
                                      (Consume_Queue, False);
                                 else
                                    Local_Host.Send_Buffer_Description
                                      (Consume_Queue, True);
                                 end if;
                                 exit;
                              exception
                                 when E : others =>
                                    Ada.Text_IO.Put_Line
                                      ("subscribe consumer exception : " &
                                         Ada.Exceptions.Exception_Name (E));
                                    delay 1.0;
                              end;
                           end loop;
                        when Subscribe_Consume_On_Fly =>
                           if Local_Host.Get_Policies (Common_Types.Consume_On_Fly) then
                              loop
                                 declare
                                    use POSIX.Message_Queues;
                                    Consume_Queue : Message_Queue_Descriptor;
                                 begin
                                    Consume_Queue := Open
                                      (POSIX.To_POSIX_String
                                         ("/consume_on_fly_" &
                                            Message.Sender),
                                       POSIX.IO.Read_Write);
                                    Queue_Vectors.Append
                                      (Consumers_On_Fly,
                                       (Queue => Consume_Queue,
                                        Uuid => Message.Sender));
                                    Local_Host.Add_Reference;
                                    Local_Host.Send_Buffer_Description
                                      (Consume_Queue, False);
                                    exit;
                                 exception
                                    when E : others =>
                                       Ada.Text_IO.Put_Line
                                         ("subscribe consumer on fly" &
                                            " exception : " &
                                            Ada.Exceptions.Exception_Name (E));
                                       delay 1.0;
                                 end;
                              end loop;
                           else
                              loop
                                 declare
                                    use POSIX.Message_Queues;
                                    Queue : Message_Queue_Descriptor;
                                    Error_Message : constant Message_Type :=
                                      (Kind => Error);
                                 begin
                                    Queue := Open
                                      (POSIX.To_POSIX_String
                                         ("/consume_on_fly_" &
                                            Message.Sender),
                                       POSIX.IO.Read_Write);
                                    Send (Queue, Error_Message);
                                    Close (Queue);
                                    exit;
                                 exception
                                    when E : others =>
                                       Ada.Text_IO.Put_Line
                                         ("subscribe consumer on fly" &
                                            " exception : " &
                                            Ada.Exceptions.Exception_Name (E));
                                       delay 1.0;
                                 end;
                              end loop;
                           end if;
                        when Subscribe_Produce =>
                           if not Producers.Is_Empty and
                             not Local_Host.Get_Policies
                             (Common_Types.Multiple_Producers)
                           then
                              declare
                                 use POSIX.Message_Queues;
                                 Queue : Message_Queue_Descriptor;
                                 Message_To_Return : constant Message_Type :=
                                   (Kind => Error);
                              begin
                                 Queue := Open
                                   (POSIX.To_POSIX_String ("/produce_" &
                                                             Message.Sender),
                                    POSIX.IO.Read_Write);
                                 Send (Queue, Message_To_Return);
                                 Close (Queue);
                              end;
                           else
                              loop
                                 declare
                                    use POSIX.Message_Queues;
                                    Produce_Queue : Message_Queue_Descriptor;
                                 begin
                                    Produce_Queue := Open
                                      (POSIX.To_POSIX_String
                                         ("/produce_" & Message.Sender),
                                       POSIX.IO.Read_Write);
                                    Local_Host.Add_Reference;
                                    Local_Host.Send_Buffer_Description
                                      (Produce_Queue, False);
                                    Producers.Append ((Queue => Produce_Queue,
                                                       Uuid => Message.Sender));
                                    exit;
                                 exception
                                    when E : others =>
                                       Ada.Text_IO.Put_Line
                                         ("subscribe producer exception : " &
                                            Ada.Exceptions.Exception_Name (E));
                                       delay 1.0;
                                 end;
                              end loop;
                           end if;
                        when Release_Free =>
                           Local_Host.Release_Free (Message, Consumers);
                        when Release_Full =>
                           Local_Host.Release_Full (Message);
                        when Release_Spy =>
                           Local_Host.Release_Spy (Message);
                        when Get_Free =>
                           declare
                              Producer : Unique_Queue_Type;
                           begin
                              Producer := Find (Producers, Message.Sender);
                              Local_Host.Append_Pending_Producer (Producer);
                           end;
                        when Get_Full =>
                           if Local_Host.Get_Policies (Common_Types.Consume_Once) then
                              declare
                                 Consumer : Unique_Queue_Type;
                              begin
                                 Consumer := Find (Consumers, Message.Sender);
                                 Local_Host.Append_Pending_Consumer (Consumer);
                              end;
                           end if;
                        when Get_Spy =>
                           if Local_Host.Get_Policies (Common_Types.Consume_On_Fly) then
                              declare
                                 Consumer : Unique_Queue_Type;
                              begin
                                 Consumer := Find (Consumers_On_Fly,
                                                   Message.Sender);
                                 Local_Host.Append_Pending_Consumer_On_Fly
                                   (Consumer);
                              end;
                           end if;
                        when Unsubscribe_Produce =>
                           declare
                              Queue : Unique_Queue_Type;
                           begin
                              Find_And_Remove (Producers, Message.Sender, Queue);
                           end;
                           Local_Host.Remove_Reference;
                        when Unsubscribe_Consume =>
                           declare
                              Queue : Unique_Queue_Type;
                           begin
                              Find_And_Remove (Consumers, Message.Sender, Queue);
                           end;
                           Local_Host.Remove_Reference;
                        when Unsubscribe_Consume_On_Fly =>
                           declare
                              Queue : Unique_Queue_Type;
                           begin
                              Find_And_Remove (Consumers_On_Fly,
                                               Message.Sender,
                                               Queue);
                           end;
                           Local_Host.Remove_Reference;
                        when New_Buffer_Available |
                          Free_Available |
                          Full_Available |
                          Spyed_Available |
                          No_More_Buffers |
                          Propagate_Delete |
                          Error =>
                           null;
                     end case;
                  end;
               exception
                  when POSIX.POSIX_Error =>
                     exit when POSIX.Get_Error_Code = POSIX.Timed_Out;
                     Ada.Text_IO.Put_Line ("posix error code" & POSIX.Get_Error_Code'Img);
                     raise;
                  when E : others =>
                     Ada.Text_IO.Put_Line ("exception name : " & Ada.Exceptions.Exception_Name (E));
                     Ada.Text_IO.Put_Line ("                 " & Ada.Exceptions.Exception_Message (E));
               end;
            end loop;
         end select;
      end loop;
   end Message_Handling_Task;

end Buffers.Shared.Host;
