with Ada.Strings.Unbounded;
with Ada.Unchecked_Deallocation;

with POSIX.IO;

package Buffers.Shared.Host is

   type Host_Type;
   type Host_Access is access all Host_Type;

   task type Message_Handling_Task is
      entry Start (Host_Item : Host_Access;
                   Timeout : Duration);
      entry Stop;
   end Message_Handling_Task;
   type Message_Handling_Access is access all Message_Handling_Task;
   procedure Free is new Ada.Unchecked_Deallocation (Message_Handling_Task,
                                                     Message_Handling_Access);

   protected type Host_Type (Messages_Handling : Message_Handling_Access) is
      new Buffer_Host_Interface with
      overriding
      procedure Set_Name (Name : String);
      overriding
      function Get_Name return String;
      overriding
      function Get_State return String;
      overriding
      function No_Data return Boolean;
      overriding
      function Compare (Name : String) return Boolean;
      overriding
      procedure Reset;
      overriding
      entry Delete;
      overriding
      function Safe_Delete return Boolean;
      overriding
      function Get_Buffers_Size return Buffer_Size_Type;
      overriding
      procedure Initialise (Buffer_Number : Positive;
                            Size : Buffer_Size_Type);
      overriding
      procedure Set_Policy (Policy : Common_Types.Policy_Type;
                            Value : Boolean);
      overriding
      procedure Set_Mode (Mode : Common_Types.Mode_Type;
                          Value : Boolean);
      --  This function gives back the policies of the buffer.
      overriding
      function Get_Policies return Policy_Array;
      --  This function gives the policies supported by the buffer.
      overriding
      function Get_Available_Policies return Policy_Array;
      overriding
      function Is_In_Error_State return Boolean;
      overriding
      entry Is_Initialised;
      function Get_Message_Task return Message_Handling_Access;
      procedure Send_Buffer_Description
        (Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
         With_Full : Boolean);
      procedure Set_On_Fly_Limit (Limit : Positive);
      procedure Add_Reference;
      procedure Remove_Reference;
      procedure Release_Free
        (Message : Message_Type;
         Consumers : Queue_Vectors.Vector);
      procedure Put_Back_In_Free_Queue
        (Buffer : in out Buffer_Description_Type);
      procedure Release_Full (Message : Message_Type);
      procedure Release_Spy  (Message : Message_Type);
      procedure Append_Pending_Consumer (Consumer : Unique_Queue_Type);
      procedure Append_Pending_Consumer_On_Fly (Consumer : Unique_Queue_Type);
      procedure Append_Pending_Producer (Producer : Unique_Queue_Type);
      function Is_Propagation_Needed return Boolean;
   private
      Need_Propagation : Boolean := False;
      Modes : Mode_Array := (Common_Types.On_Fly_Refresh => True,
                             Common_Types.No_Transmit => False);
      Policies : Policy_Array := (others => False);
      Name : Ada.Strings.Unbounded.Unbounded_String :=
        Ada.Strings.Unbounded.Null_Unbounded_String;
      Buffer_Size : Buffer_Size_Type := 0;
      Free_Buffers : Vector;
      Full_Buffers : Vector;
      Spyed_Buffers : Vector;
      Ordered_Buffers : Vector;
      Pending_Consumers : Queue_Vectors.Vector;
      Pending_Consumers_On_Fly : Queue_Vectors.Vector;
      Pending_Producers : Queue_Vectors.Vector;
      Initialised : Boolean := False;
      Memory_Descriptor : POSIX.IO.File_Descriptor;
      Initial_Depth : Positive := 1;
      Nb_Ref : Natural := 0;
      Spyed_Limit : Natural := 0;
   end Host_Type;

   procedure Free is new Ada.Unchecked_Deallocation (Host_Type, Host_Access);

   Incompatible_Call_With_Policy : exception;

private

   Available_Policies : constant Policy_Array :=
     (Common_Types.Multiple_Producers => True,
      Common_Types.Consume_Once => True,
      Common_Types.Consume_On_Fly => True);

end Buffers.Shared.Host;
