with Ada.Exceptions;
with Ada.Text_IO;
with System.Storage_Elements;

with POSIX.Permissions;
with POSIX.Shared_Memory_Objects;
with POSIX.Memory_Mapping;

with Uuid;

package body Buffers.Shared.Produce is

   use Ada.Strings.Unbounded;

   ------------------
   -- Produce_Type --
   ------------------

   protected body Produce_Type is

      procedure Set_Host_Available (New_Value : Boolean) is
      begin
         Host_Available := New_Value;
      end Set_Host_Available;

      ---------------------
      -- Get_Free_Buffer --
      ---------------------

      entry Get_Free_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      when not Ask_Buffer_Done or
        not Is_Empty (Free_Buffers) or
         not Host_Available is
         Buffer : Buffer_Description_Type;
      begin
         if not Host_Available then
            raise Host_Not_Available with "get_free_buffer need an available host";
         end if;
         if Buffer_Handle.State /= Raw then
            raise Invalid_Buffer_State with
              Buffer_Handle.State'Img & " need a raw handle";
         end if;
         if Ask_Buffer_Done then
            Buffer_In_Use := True;
            Buffer := First_Element (Free_Buffers);
            Delete_First (Free_Buffers);
            Buffer_Handle.State := In_Use_Filling;
            Buffer_Handle.Address := Buffer.Address;
            Buffer_Handle.Bytes_Available := Buffer.Bytes;
            Buffer_Handle.Bytes_Used := 0;
            Buffer_Handle.Buffer_ID := Buffer.ID;
            if Free_Buffers.Is_Empty then
               Ask_Buffer_Done := False;
            end if;
         else
            Send (Server_Queue,
                  (Kind => Get_Free,
                   Sender => Uuid_Image));
            Ask_Buffer_Done := True;
            requeue Get_Free_Buffer with abort;
         end if;
      end Get_Free_Buffer;

      entry Block_Free when not Is_Empty (Free_Buffers) is
      begin
         null;
      end Block_Free;

      -------------------------
      -- Release_Free_Buffer --
      -------------------------

      procedure Release_Free_Buffer
        (Buffer_Handle : in out Buffer_Handle_Type'Class)
      is
         use type Common_Types.Buffer_Size_Type;
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Dumping =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Filling =>
               if Buffer_Handle.Bytes_Used = 0 then
                  raise Invalid_Call;
               end if;
               Send (Server_Queue,
                     (Kind => Release_Free,
                      ID => Buffer_Handle.Buffer_ID,
                      Bytes => Buffer_Handle.Bytes_Used));
         end case;
         Buffer_Handle.State := Clean;
         Buffer_In_Use := False;
      end Release_Free_Buffer;

      procedure Release_Free_Buffer_Not_Used
        (Buffer_Handle : in out Buffer_Handle_Type'Class) is
         Buffer : Buffer_Description_Type;
      begin
         case Buffer_Handle.State is
            when Raw | Clean | In_Use_Dumping =>
               raise Invalid_Buffer_State
                 with Buffer_Handle.State'Img &
                 " isn't a valid state for release";
            when In_Use_Filling =>
               Buffer := (Address => Buffer_Handle.Address,
                          Bytes => Buffer_Size,
                          ID => Buffer_Handle.Buffer_ID,
                          Nb_Ref => 0);
               Free_Buffers.Append (Buffer);
         end case;
         Buffer_Handle.State := Clean;
         Buffer_In_Use := False;
      end Release_Free_Buffer_Not_Used;

      function Free_Is_Empty return Boolean is
      begin
         return Is_Empty (Free_Buffers);
      end Free_Is_Empty;

      --------------
      -- Set_Name --
      --------------

      procedure Set_Name (Name : String) is
         Uuid_Value : Uuid.Uuid_Type;
         use POSIX.Message_Queues;
      begin
         if Produce_Type.Name /= Null_Unbounded_String then
            raise Already_Initialised with "name is already set : " &
              To_String (Produce_Type.Name);
         end if;
         Produce_Type.Name := To_Unbounded_String (Name);
         Uuid.Uuid_Generate (Uuid_Value);
         Uuid_Image := Uuid.Image (Uuid_Value);
         Server_Queue := Open (POSIX.To_POSIX_String ("/" & Name),
                               POSIX.IO.Read_Write);
         Memory_Descriptor := POSIX.Shared_Memory_Objects.Open_Shared_Memory
           (POSIX.To_POSIX_String ("/" & Name),
            POSIX.IO.Read_Write);
         Send (Server_Queue,
               (Kind => Subscribe_Produce,
                Sender => Uuid_Image));
         Host_Available := True;
      end Set_Name;

      function Get_Name return String is
      begin
         return To_String (Name);
      end Get_Name;

      function Get_State return String is
      begin
         return "<buffer type=""local"" error=""" & Error_State'Img & """>" &
           "<name>" & To_String (Name) & "</name>" &
           "<size>" & Buffer_Size'Img & "</size>" &
           "<depth>" & Initial_Depth'Img & "</depth>" &
           "<free>" & Free_Buffers.Length'Img & "</free>" &
           "</buffer>";
      end Get_State;

      function No_Data return Boolean is
      begin
         return not Buffer_In_Use;
      end No_Data;

      function Free_Occupancy_Rate return Natural is
      begin
         return Natural (Free_Buffers.Length);
      end Free_Occupancy_Rate;

      function Free_Depth return Positive is
      begin
         return Initial_Depth;
      end Free_Depth;

      function Compare (Name : String) return Boolean is
      begin
         return Name = Produce_Type.Name;
      end Compare;

      -----------
      -- Reset --
      -----------

      function Safe_Delete return Boolean is
      begin
         if Buffer_In_Use then
            return False;
         end if;
         if not Free_Buffers.Is_Empty then
            return False;
         end if;
         return True;
      end Safe_Delete;

      procedure Delete is
      begin
         for Buffer of Buffers loop
            POSIX.Memory_Mapping.Unmap_Memory
              (Buffer.Address,
               System.Storage_Elements.Storage_Offset
               (Buffer.Bytes));
         end loop;
         POSIX.IO.Close (Memory_Descriptor);
         Clear (Free_Buffers);
         Clear (Buffers);
         POSIX.Message_Queues.Unlink_Message_Queue
           (POSIX.To_POSIX_String ("/produce_" & Uuid_Image));
         Send (Server_Queue,
               (Kind => Unsubscribe_Produce,
                Sender => Uuid_Image));
      end Delete;

      procedure Reset is
      begin
         if Buffer_In_Use then
            raise Cant_Reset;
         end if;
         if not Free_Buffers.Is_Empty then
            raise Cant_Reset;
         end if;
      end Reset;

      ----------------------
      -- Get_Buffers_Size --
      ----------------------

      function Get_Buffers_Size return Buffer_Size_Type is
      begin
         return Buffer_Size;
      end Get_Buffers_Size;

      --------------------
      -- Free_Available --
      --------------------

      procedure Free_Available (Message : Message_Type) is
         Buffer : Buffer_Description_Type;
      begin
         Buffer := Find (Buffers, Message.ID);
         Free_Buffers.Append ((Address => Buffer.Address,
                               Bytes => Buffer_Size,
                               ID => Message.ID,
                               Nb_Ref => 0));
      end Free_Available;

      function Is_In_Error_State return Boolean is
      begin
         return Error_State;
      end Is_In_Error_State;

      entry Is_Initialised when Initialised is
      begin
         null;
      end Is_Initialised;

      procedure No_More_Buffers_Message is
      begin
         Initialised := True;
      end No_More_Buffers_Message;

      procedure Append_Buffer (Buffer : Buffer_Description_Type) is
      begin
         Buffers.Append (Buffer);
         Initial_Depth := Initial_Depth + 1;
      end Append_Buffer;

      function Get_Uuid_Image return String is
      begin
         return Uuid_Image;
      end Get_Uuid_Image;

      procedure Set_Size (Size : Buffer_Size_Type) is
      begin
         Buffer_Size := Size;
      end Set_Size;

      function Get_Memory_Descriptor return POSIX.IO.File_Descriptor is
      begin
         return Memory_Descriptor;
      end Get_Memory_Descriptor;

      procedure Error_Message is
      begin
         Error_State := True;
      end Error_Message;

   end Produce_Type;

   task body Message_Handling_Task is
      Message_Check_Timeout : Duration;
      Produce_Queue : POSIX.Message_Queues.Message_Queue_Descriptor;
      Address : System.Address;
      Attrs : POSIX.Message_Queues.Attributes;
      use POSIX.Message_Queues;
      Offset : POSIX.IO_Count := 0;
   begin
      select
         accept Start (Timeout : Duration) do
            Message_Check_Timeout := Timeout;
            POSIX.Message_Queues.Set_Max_Messages (Attrs, 64);
            POSIX.Message_Queues.Set_Message_Length (Attrs, Message_Size);
            Produce_Queue := Open_Or_Create
              (POSIX.To_POSIX_String ("/produce_" &
                                        Wrapper.Producer.Get_Uuid_Image),
               POSIX.IO.Read_Write,
               POSIX.Permissions.Owner_Permission_Set,
               Attrs => Attrs);
         end Start;
      or
         terminate;
      end select;
      loop
         select
            accept Stop;
            exit;
         else
            begin
               declare
                  Message : constant Message_Type := Relative_Timed_Receive (Produce_Queue,
                                                                             Message_Check_Timeout);
                  use type POSIX.IO_Count;
                  use type POSIX.Memory_Mapping.Protection_Options;
                  use type Common_Types.Buffer_Size_Type;
               begin
                  case Message.Kind is
                     when Propagate_Delete =>
                        Wrapper.Producer.Set_Host_Available (False);
                     when Free_Available =>
                        Wrapper.Producer.Free_Available (Message);
                     when New_Buffer_Available =>
                        if Wrapper.Producer.Get_Buffers_Size = 0 then
                           Wrapper.Producer.Set_Size (Message.Bytes);
                        end if;
                        Address := POSIX.Memory_Mapping.Map_Memory
                          (System.Storage_Elements.Storage_Offset (Message.Bytes),
                           POSIX.Memory_Mapping.Allow_Read +
                             POSIX.Memory_Mapping.Allow_Write,
                           POSIX.Memory_Mapping.Map_Shared,
                           Wrapper.Producer.Get_Memory_Descriptor,
                           Offset);
                        Offset := Offset + POSIX.IO_Count (Message.Bytes);
                        Wrapper.Producer.Append_Buffer
                          ((Address => Address,
                            Bytes => Message.Bytes,
                            ID => Message.ID,
                            Nb_Ref => 0));
                     when No_More_Buffers =>
                        Wrapper.Producer.No_More_Buffers_Message;
                     when Error =>
                        Wrapper.Producer.Error_Message;
                     when Subscribe_Consume .. Release_Spy |
                       Full_Available .. Unsubscribe_Consume_On_Fly =>
                        null;
                  end case;
               end;
            exception
               when POSIX.POSIX_Error =>
                  null;
               when E : others =>
                  Ada.Text_IO.Put_Line ("unexpected exception : " &
                                          Ada.Exceptions.Exception_Name (E));
            end;
         end select;
      end loop;
   end Message_Handling_Task;

   procedure Clean (Vector : in out Producer_Vectors.Vector) is
   begin
      for Couple of Vector loop
         Free (Couple);
      end loop;
      Vector.Clear;
   end Clean;

end Buffers.Shared.Produce;
