with Ada.Exceptions;
with Ada.Text_IO;

package body Buffers.Shell is

   use Ada.Strings.Unbounded;

   ------------------
   -- Buffer_Exist --
   ------------------

   function Buffer_Exist
     (Buffer_Container_Vector : Buffer_Container_Vectors.Vector;
      Name : String)
      return Boolean
   is
   begin
      return Get_Buffer_Container (Buffer_Container_Vector, Name) /= null;
   end Buffer_Exist;

   ----------------------------
   --  Get_Buffer_Container  --
   ----------------------------

   function Get_Buffer_Container
     (Buffer_Container_Vector : Buffer_Container_Vectors.Vector;
      Name : String) return Buffer_Container_Access is
   begin
      for Buffer_Container of Buffer_Container_Vector loop
         if Buffer_Container.Name = Name then
            return Buffer_Container;
         end if;
      end loop;
      return null;
   end Get_Buffer_Container;

   --------------------
   -- Memory_Handler --
   --------------------

   protected body Memory_Handler is

      ---------------
      -- Get_State --
      ---------------

      function Get_State return String is
      begin
         return "<error cmd=""get_state"">not implemented</error>";
      end Get_State;

      ----------------------
      -- Get_Buffer_State --
      ----------------------

      function Get_Buffer_State (Name : String) return String is
      begin
         if not Buffer_Exist (Buffers_Vector, Name) then
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         for Buffer_Container of Buffers_Vector loop
            if Name = Buffer_Container.Name then
               return Buffer_Container.Buffer.Get_State;
            end if;
         end loop;
         return "<error cmd=""get_buffer_state"" name=""" & Name &
           """>not found</error>";
      end Get_Buffer_State;

      function Get_Buffer_Size (Name : String) return Buffer_Size_Type is
      begin
         if not Buffer_Exist (Buffers_Vector, Name) then
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         for Buffer_Container of Buffers_Vector loop
            if Name = Buffer_Container.Name then
               return Buffer_Container.Buffer.Get_Buffers_Size;
            end if;
         end loop;
         return 0;
      end Get_Buffer_Size;

      -------------------
      -- Create_Buffer --
      -------------------

      procedure Create_Buffer
        (Name : String;
         Buffer_Container : out Buffer_Container_Access;
         Size : Common_Types.Buffer_Size_Type := 65_536;
         Depth : Positive := 3)
      is
         Message_Handling : Buffers.Shared.Host.Message_Handling_Access;
      begin
         --  Ada.Text_IO.Put_Line ("create buffer : " & Name &
         --                          Size'Img & Depth'Img);
         if Buffer_Exist (Buffers_Vector, Name) then
            raise Shell_Error with
              "buffer named " & Name & " already exists";
         end if;
         Message_Handling := new Buffers.Shared.Host.Message_Handling_Task;
         Buffer_Container := new Buffer_Container_Type (Message_Handling);
         Buffer_Container.Buffer.Set_Name (Name);
         --  pragma Warnings (Off);
         --  Message_Handling.Start (Buffer_Container.Buffer'Access, 1.0);
         --  pragma Warnings (On);
         Buffer_Container.Name := To_Unbounded_String (Name);
         Buffer_Container.Buffer.Initialise (Buffer_Number => Depth,
                                             Size => Size);
         Buffers_Vector.Append (Buffer_Container);
      end Create_Buffer;

      ------------------
      -- Reset_Buffer --
      ------------------

      procedure Reset_Buffer (Name : String) is
      begin
         --  Ada.Text_IO.Put_Line ("reset buffer step 0 : " & Name);
         if not Buffer_Exist (Buffers_Vector, Name) then
            --  Ada.Text_IO.Put_Line ("reset buffer " & Name &
            --  "impossible, buffer isn't in the list");
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         --  Ada.Text_IO.Put_Line ("reset buffer step 1 ");
         for Buffer_Container of Buffers_Vector loop
            --  Ada.Text_IO.Put_Line ("reset buffer step 2");
            if Buffer_Container.Name = Name then
               --  Ada.Text_IO.Put_Line ("reset buffer step 3");
               Buffer_Container.Buffer.Reset;
               exit;
            end if;
         end loop;
      exception
         when E : others =>
            Ada.Text_IO.Put_Line ("reset buffer exception : " &
                                    Ada.Exceptions.Exception_Name (E));
            raise;
      end Reset_Buffer;

      -------------------
      -- Delete_Buffer --
      -------------------

      procedure Delete_Buffer (Name : String) is
         Cursor : Buffer_Container_Vectors.Cursor;
         use type Buffer_Container_Vectors.Cursor;
      begin
         --  Ada.Text_IO.Put_Line ("delete buffer step 0 : " & Name);
         if not Buffer_Exist (Buffers_Vector, Name) then
            --  Ada.Text_IO.Put_Line ("delete buffer " & Name &
            --  "impossible, buffer isn't in the list");
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         --  Ada.Text_IO.Put_Line ("delete buffer step 1 ");
         Cursor := Buffers_Vector.First;
         loop
            exit when Cursor = Buffer_Container_Vectors.No_Element;
            --  Ada.Text_IO.Put_Line ("delete buffer step 1");
            declare
               Buffer_Container : Buffer_Container_Access :=
                 Buffer_Container_Vectors.Element (Cursor);
               Task_Access : Buffers.Shared.Host.Message_Handling_Access;
            begin
               if Buffer_Container.Name = Name then
                  --  Ada.Text_IO.Put_Line ("delete buffer step 2 ");
                  --  Ada.Text_IO.Put_Line ("delete buffer step 3 ");
                  pragma Warnings (Off);
                  Buffer_Container.Buffer.Delete;
                  Buffer_Container.Message_Handling.Stop;
                  pragma Warnings (On);
                  --  Ada.Text_IO.Put_Line ("delete buffer step 4 ");
                  Task_Access := Buffer_Container.Message_Handling;
                  Buffers.Shared.Host.Free (Task_Access);
                  Free (Buffer_Container);
                  Buffers_Vector.Delete (Cursor);
                  exit;
               else
                  Cursor := Buffer_Container_Vectors.Next (Cursor);
               end if;
            end;
         end loop;
      exception
         when E : others =>
            Ada.Text_IO.Put_Line ("delete buffer exception : " &
                                    Ada.Exceptions.Exception_Name (E));
            raise;
      end Delete_Buffer;

      ----------------
      -- Set_Policy --
      ----------------

      procedure Set_Policy
        (Name : String;
         Policy : Common_Types.Policy_Type;
         Value : Boolean)
      is
      begin
         if not Buffer_Exist (Buffers_Vector, Name) then
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         for Buffer_Container of Buffers_Vector loop
            if Name = Buffer_Container.Name then
               Buffer_Container.Buffer.Set_Policy (Policy, Value);
            end if;
         end loop;
      end Set_Policy;

      --------------
      -- Set_Mode --
      --------------

      procedure Set_Mode
        (Name : String;
         Mode : Common_Types.Mode_Type;
         Value : Boolean)
      is
      begin
         if not Buffer_Exist (Buffers_Vector, Name) then
            raise Shell_Error with Name & " buffer isn't present in PMH";
         end if;
         for Buffer_Container of Buffers_Vector loop
            if Name = Buffer_Container.Name then
               Buffer_Container.Buffer.Set_Mode (Mode, Value);
            end if;
         end loop;
      end Set_Mode;

      ----------------------
      -- Get_Buffer_Names --
      ----------------------

      function Get_Buffer_Names return Common_Types.String_Vectors.Vector is
         Names : Common_Types.String_Vectors.Vector;
      begin
         for Buffer_Container of Buffers_Vector loop
            Names.Append (To_String (Buffer_Container.Name));
         end loop;
         return Names;
      end Get_Buffer_Names;

      function Buffer_Exist (Name : String) return Boolean is
      begin
         return Buffer_Exist (Buffers_Vector, Name);
      end Buffer_Exist;

   end Memory_Handler;

end Buffers.Shell;
