package body Buffers is

   ----------
   -- Free --
   ----------

   procedure Free (Buffer_Handle : in out Buffer_Handle_Access) is
   begin
      Private_Handle_Free (Buffer_Handle);
      Buffer_Handle := null;
   end Free;

   -------------------------
   -- Get_Available_Bytes --
   -------------------------

   function Get_Available_Bytes (Buffer_Handle : Buffer_Handle_Type)
                      return Buffer_Size_Type is
   begin
      if Buffer_Handle.State /= In_Use_Filling then
         raise Only_Used_On_Free_Buffers;
      end if;
      return Buffer_Handle.Bytes_Available;
   end Get_Available_Bytes;

   --------------------
   -- Get_Used_Bytes --
   --------------------

   function Get_Used_Bytes (Buffer_Handle : Buffer_Handle_Type)
                      return Buffer_Size_Type is
   begin
      if Buffer_Handle.State /= In_Use_Dumping then
         raise Only_Used_On_Full_Buffers;
      end if;
      return Buffer_Handle.Bytes_Used;
   end Get_Used_Bytes;

   --------------------
   -- Set_Used_Bytes --
   --------------------

   procedure Set_Used_Bytes (Buffer_Handle : in out Buffer_Handle_Type;
                             Bytes : Buffer_Size_Type) is
   begin
      if Buffer_Handle.State /= In_Use_Filling then
         raise Only_Used_On_Free_Buffers;
      end if;
      Buffer_Handle.Bytes_Used := Bytes;
   end Set_Used_Bytes;

   -----------------
   -- Get_Address --
   -----------------

   function Get_Address
     (Buffer_Handle : Buffer_Handle_Type)
      return System.Address
   is
   begin
      return Buffer_Handle.Address;
   end Get_Address;

   procedure Reuse (Buffer_Handle : in out Buffer_Handle_Type) is
   begin
      case Buffer_Handle.State is
         when Clean | Raw =>
            Buffer_Handle.State := Raw;
         when others =>
            raise Buffer_Handle_Misuse with
              Buffer_Handle.State'Img &
              " buffer handle must be clean or raw to be reused";
      end case;
   end Reuse;

   --------------
   -- Finalize --
   --------------

   procedure Finalize (Buffer_Handle : in out Buffer_Handle_Type) is
   begin
      case Buffer_Handle.State is
         when Clean | Raw =>
            null;
         when others =>
            raise Buffer_Handle_Misuse with
              Buffer_Handle.State'Img &
              " buffer handle may be misused in your code";
      end case;
   end Finalize;

   function Find (Buffers_Vector : Vector;
                  ID : Buffer_ID_Type) return Buffer_Description_Type is
   begin
      for Buffer_Description of Buffers_Vector loop
         if Buffer_Description.ID = ID then
            return Buffer_Description;
         end if;
      end loop;
      raise ID_Not_Found with ID'Img & " not found";
   end Find;

   procedure Find_And_Remove (Buffers_Vector : in out Vector;
                              ID : Buffer_ID_Type;
                              Buffer : out Buffer_Description_Type) is
      L_Cursor : Buffer_Description_Vectors.Cursor;
   begin
      L_Cursor := First (Buffers_Vector);
      loop
         exit when L_Cursor = No_Element;
         if Element (L_Cursor).ID = ID then
            Buffer := Element (L_Cursor);
            Delete (Buffers_Vector, L_Cursor);
            return;
         end if;
         L_Cursor := Next (L_Cursor);
      end loop;
      raise ID_Not_Found with ID'Img & " not found";
   end Find_And_Remove;

   function Is_In (Buffers_Vector : Vector;
                   ID : Buffer_ID_Type) return Boolean is
   begin
      for Buffer_Description of Buffers_Vector loop
         if Buffer_Description.ID = ID then
            return True;
         end if;
      end loop;
      return False;
   end Is_In;

   package body Generic_Buffers is

      function To_Word_Count (Size : Buffer_Size_Type) return Natural is
         use type Common_Types.Buffer_Size_Type;
      begin
         return Natural (Size / (Element_Type'Size / 8));
      end To_Word_Count;

      function To_Bytes (Number_Of_Word : Natural) return Buffer_Size_Type is
      begin
         return Buffer_Size_Type ((Element_Type'Size / 8) * Number_Of_Word);
      end To_Bytes;

   end Generic_Buffers;

   function Compare (Buffer : Buffer_Produce_Access;
                     Name : String) return Boolean is
   begin
      return Compare (Buffer.all, Name);
   end Compare;

   function Compare (Buffer : Buffer_Consume_Access;
                     Name : String) return Boolean is
   begin
      return Compare (Buffer.all, Name);
   end Compare;

end Buffers;
