with Ada.Finalization;
with Ada.Unchecked_Deallocation;
with Ada.Containers.Vectors;
with Interfaces;
with System;

with Common_Types;

package Buffers is
   pragma Preelaborate;
   subtype Buffer_Size_Type is Common_Types.Buffer_Size_Type;
   type Buffer_Handle_Type is tagged private;
   type Buffer_Handle_Access is access all Buffer_Handle_Type;
   --  This procedure release the memory allocated by the new
   --  operator.
   procedure Free (Buffer_Handle : in out Buffer_Handle_Access);

   --  This function returns the number of bytes one can put in the buffer
   --  handled by the buffer_handle. It will raise Only_Used_On_Free_Buffers
   --  if called on a buffer_handle connected to a full buffer.
   function Get_Available_Bytes (Buffer_Handle : Buffer_Handle_Type)
                                return Buffer_Size_Type;
   --  This function returns the number of bytes already in the full buffer.
   --  It will raise Only_Used_On_Full_Buffers if called on a buffer_handle
   --  connected to a free buffer.
   function Get_Used_Bytes (Buffer_Handle : Buffer_Handle_Type)
                           return Buffer_Size_Type;
   --  This procedure sets the number of bytes contained in a free buffer
   --  in filling mode just before releasing it. It will raise
   --  Only_Used_On_Free_Buffers if called on a buffer_handle connected to a
   --  full buffer.
   procedure Set_Used_Bytes (Buffer_Handle : in out Buffer_Handle_Type;
                             Bytes : Buffer_Size_Type);
   --  This function return the system address where to put data if the
   --  buffer is a free one of where to get the data if the buffer is a full
   --  one.
   function Get_Address (Buffer_Handle : Buffer_Handle_Type)
                        return System.Address;
   --  One a buffer_handle is used, it should terminate its life cycle as
   --  clean. This procedure change the buffer_handle from clean to raw is
   --  the buffer_handle was already used. If the state isn't raw or clean the
   --  procedure will raise a Buffer_Handle_Misuse exception.
   procedure Reuse (Buffer_Handle : in out Buffer_Handle_Type);

   Only_Used_On_Free_Buffers : exception;
   Only_Used_On_Full_Buffers : exception;
   Invalid_Buffer_State : exception;
   Invalid_Call : exception;
   Buffer_Handle_Misuse : exception;
   Cant_Reset : exception;

   --  All buffers should have properties in common: name, reset, size and
   --  number.
   type Buffer_Common_Interface is synchronized interface;
   --  This procedure sets the name of a buffer.
   procedure Set_Name (Buffer : in out Buffer_Common_Interface;
                       Name : String) is abstract;
   --  This function returns the name of a buffer.
   function Get_Name (Buffer : Buffer_Common_Interface)
                     return String is abstract;
   --  This function returns true if the buffer has the same name as the
   --  string parameter Name.
   function Compare (Buffer : Buffer_Common_Interface;
                     Name : String) return Boolean is abstract;
   --  This procedure resets the buffer.
   procedure Reset (Buffer : in out Buffer_Common_Interface) is abstract;
   --  This procedure deletes the buffer.
   procedure Delete (Buffer : in out Buffer_Common_Interface) is abstract;
   --  This function returns true if the buffer can be deleted in a safe way.
   --  Mainly when there isn't any full buffers waiting to be consumed.
   function Safe_Delete (Buffer : Buffer_Common_Interface) return Boolean
      is abstract;
   --  This function returns the number of bytes a buffer's bloc can contain.
   function Get_Buffers_Size (Buffer : Buffer_Common_Interface)
                             return Buffer_Size_Type is abstract;
   --  This function returns the occupancy of the Free and Full queues as a xml
   --  string.
   function Get_State (Buffer : Buffer_Common_Interface)
                      return String is abstract;
   --  This function returns true if there isn't any data in the buffer nether
   --  any buffer beeing filled at the calling time.
   function No_Data (Buffer : Buffer_Common_Interface)
                    return Boolean is abstract;
   --  This function returns true if the object is in error.
   function Is_In_Error_State (Buffer : Buffer_Common_Interface)
                              return Boolean is abstract;
   procedure Is_Initialised (Buffer : in out Buffer_Common_Interface)
      is abstract;

   type Buffer_Consume_Interface is synchronized interface and
     Buffer_Common_Interface;
   --  This entry should block until a full buffer is available.
   procedure Block_Full (Buffer : in out Buffer_Consume_Interface) is abstract;
   --  This procedure gets a full buffer if possible (will block if no full
   --  buffer available).
   procedure Get_Full_Buffer
     (Buffer : in out Buffer_Consume_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  Once a full buffer is of no more use, one can release it using this
   --  procedure.
   procedure Release_Full_Buffer
     (Buffer : in out Buffer_Consume_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  If the full buffer wasn't used one can put it back in the full buffer
   --  queue.
   procedure Release_Full_Buffer_Not_Used
     (Buffer : in out Buffer_Consume_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  This function returns true if the full queue is empty.
   function Full_Is_Empty (Buffer : Buffer_Consume_Interface)
                          return Boolean is abstract;
   --  This function returns the number of bytes available in the first buffer
   --  in the full queue. If there isn't any full buffer it returns 0.
   function Full_Size (Buffer : Buffer_Consume_Interface)
                      return Buffer_Size_Type is abstract;

   type Buffer_Consume_Access is access all Buffer_Consume_Interface'Class;
   --  This function returns true if the buffer has the same name as the
   --  string parameter Name.
   function Compare (Buffer : Buffer_Consume_Access;
                     Name : String) return Boolean;
   procedure Free is new Ada.Unchecked_Deallocation
     (Buffer_Consume_Interface'Class, Buffer_Consume_Access);
   type Buffer_Consume_Array is array (Positive range <>) of
     Buffer_Consume_Access;
   package Buffer_Consume_Vectors is new Ada.Containers.Vectors
     (Positive, Buffer_Consume_Access, "=");

   type Buffer_Produce_Interface is synchronized interface and
     Buffer_Common_Interface;
   --  This entry should block until a free buffer is available.
   procedure Block_Free (Buffer : in out Buffer_Produce_Interface) is abstract;
   --  This procedure gets a free buffer if possible (will block if no free
   --  buffer available).
   procedure Get_Free_Buffer
     (Buffer : in out Buffer_Produce_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  Once a free buffer is filled with data, one can release it using this
   --  procedure.
   procedure Release_Free_Buffer
     (Buffer : in out Buffer_Produce_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  If the buffer wasn't filled one can put it back in the free buffer
   --  queue.
   procedure Release_Free_Buffer_Not_Used
     (Buffer : in out Buffer_Produce_Interface;
      Buffer_Handle : in out Buffer_Handle_Type'Class) is abstract;
   --  This function returns true if the free queue is empty.
   function Free_Is_Empty (Buffer : Buffer_Produce_Interface)
                          return Boolean is abstract;
   --  This function returns the number of free blocs in the free queue.
   function Free_Occupancy_Rate (Buffer : Buffer_Produce_Interface)
                                return Natural is abstract;
   --  This function returns the number of free buffers available after
   --  initialisation.
   function Free_Depth (Buffer : Buffer_Produce_Interface)
                       return Positive is abstract;

   type Buffer_Produce_Access is access all Buffer_Produce_Interface'Class;
   --  This function returns true if the buffer has the same name as the
   --  string parameter Name.
   function Compare (Buffer : Buffer_Produce_Access;
                     Name : String) return Boolean;
   procedure Free is new Ada.Unchecked_Deallocation
     (Buffer_Produce_Interface'Class, Buffer_Produce_Access);
   type Buffer_Produce_Array is array (Positive range <>) of
     Buffer_Produce_Access;
   package Buffer_Produce_Vectors is new Ada.Containers.Vectors
     (Positive, Buffer_Produce_Access, "=");

   type Policy_Array is array (Common_Types.Policy_Type) of Boolean;
   Not_Supported_Policy : exception;
   type Mode_Array is array (Common_Types.Mode_Type) of Boolean;

   type Buffer_Host_Interface is synchronized interface and
     Buffer_Common_Interface;
   --  This procedure allocates Buffer_Number blocks of memory with a number
   --  of bytes Size.
   procedure Initialise (Buffer : in out Buffer_Host_Interface;
                         Buffer_Number : Positive;
                         Size : Buffer_Size_Type) is abstract;
   --  This procedure should be used to enable/disable a buffer policy.
   procedure Set_Policy (Buffer : in out Buffer_Host_Interface;
                         Policy : Common_Types.Policy_Type;
                         Value : Boolean) is abstract;
   --  This procedure should be used to enable/disable a buffer mode.
   procedure Set_Mode (Buffer : in out Buffer_Host_Interface;
                       Mode : Common_Types.Mode_Type;
                       Value : Boolean) is abstract;
   --  This function gives back the policies of the buffer.
   function Get_Policies (Buffer : Buffer_Host_Interface) return Policy_Array
      is abstract;
   --  This function gives the policies supported by the buffer.
   function Get_Available_Policies (Buffer : Buffer_Host_Interface)
                                   return Policy_Array is abstract;

   type Buffer_ID_Type is new Interfaces.Unsigned_32;
   type Buffer_Description_Type is record
      Address : System.Address;
      Bytes : Buffer_Size_Type;
      ID : Buffer_ID_Type;
      Nb_Ref : Natural;
   end record;
   package Buffer_Description_Vectors is new Ada.Containers.Vectors
     (Positive, Buffer_Description_Type);
   use Buffer_Description_Vectors;
   --  This funtion returns the buffer descripton of a buffer given the buffer
   --  uuid. It will raise Uuid_Not_Found if no match.
   function Find (Buffers_Vector : Vector;
                  ID : Buffer_ID_Type) return Buffer_Description_Type;
   --  This procedure find and remove the buffer descripton of a buffer given
   --  the buffer uuid. It will raise Uuid_Not_Found if no match.
   procedure Find_And_Remove (Buffers_Vector : in out Vector;
                              ID : Buffer_ID_Type;
                              Buffer : out Buffer_Description_Type);
   --  This function returns true if a buffer with this uuid is in the vector.
   function Is_In (Buffers_Vector : Vector;
                   ID : Buffer_ID_Type) return Boolean;
   ID_Not_Found : exception;

   generic
      type Element_Type is private;
   package Generic_Buffers is

      type Element_Array is array (Positive range <>) of Element_Type;
      type Element_Array_Access is access all Element_Array;

      --  This function converts from Buffer_Size_Type to Natural.
      function To_Word_Count (Size : Buffer_Size_Type) return Natural;
      --  This function converts from Natural to Buffer_Size_Type.
      function To_Bytes (Number_Of_Word : Natural) return Buffer_Size_Type;

   end Generic_Buffers;

private
   type State_Type is (Raw, Clean, In_Use_Filling, In_Use_Dumping);
   type Buffer_Handle_Type is new Ada.Finalization.Controlled
     with record
      State : State_Type := Raw;
      Bytes_Available : Buffer_Size_Type;
      Bytes_Used : Buffer_Size_Type;
      Address : System.Address;
      Buffer_ID : Buffer_ID_Type;
   end record;
   --  This procedure ensure the good use of the buffer handler. Handle state
   --  has to be Raw or Clean, if not a Buffer_Handle_Misuse is raised. This
   --  make sure no free or full buffers are left unreleased when going out of
   --  the buffer handle scope.
   overriding
   procedure Finalize (Buffer_Handle : in out Buffer_Handle_Type);

   procedure Private_Handle_Free is new Ada.Unchecked_Deallocation
     (Buffer_Handle_Type, Buffer_Handle_Access);

end Buffers;
