with Ada.Exceptions;
with Ada.Text_IO;

with GNAT.Sockets;

with Buffers.Shell;
with Broadcast_Stuff;
with Broadcast_Buffers;
with Pmh_Directory;

with Global_Options;

with Dcod_Ctl_Service.Client;

package body DCOD.Pmh.Real_Handler is

   use Ada.Strings.Unbounded;

   procedure Callback (Ressource : in out Broadcast_Stuff.Buffer_Ressource_Type;
                       Address : GNAT.Sockets.Sock_Addr_Type);

   procedure Callback (Ressource : in out Broadcast_Stuff.Buffer_Ressource_Type;
                       Address : GNAT.Sockets.Sock_Addr_Type) is
      Ip_String : constant String := GNAT.Sockets.Image (Address.Addr);
   begin
      Ada.Text_IO.Put_Line ("Callback for " & Ressource.Get_Name & " from " & Ip_String);
      declare
         Local_Ctl : constant String :=
           Broadcast_Stuff.Get_Process (Pmh_Directory.Directory,
                                        Broadcast_Stuff.Common_Transport_Layer,
                                        Global_Options.Get_Name,
                                        GNAT.Sockets.Host_Name);
         Buffer_Name : constant String := To_String (Ressource.Name);
      begin
         Dcod_Ctl_Service.Client.Request_Server ("TL://" & Buffer_Name,
                                                 Buffer_Name,
                                                 Local_Ctl);
         Ressource.Served := True;
         Ada.Text_IO.Put_Line ("local output declaration done");
      end;
   exception
      when Broadcast_Stuff.Process_Not_Found =>
         Ressource.Error := True;
   end Callback;

   task body Memory_Handler_Task is
      Container : Buffers.Shell.Buffer_Container_Access;
      use type Buffers.Shell.Buffer_Container_Access;
      Serve_Remotly : Boolean := False;
      Ressource_Name : Ada.Strings.Unbounded.Unbounded_String;
      Watchdog : Natural := 0;
   begin
      Broadcast_Buffers.Ressource_Shell.Subscribe (Callback'Access);
      loop
         select
            accept Provide_Buffer
              (Name : String;
               Size : Common_Types.Buffer_Size_Type := 65_536;
               Depth : Positive := 3) do
               Buffers.Shell.Memory_Handler.Create_Buffer (Name,
                                                           Container,
                                                           Size,
                                                           Depth);
               Container.Message_Handling.Start (Container.Buffer'Access, 1.0);
               Broadcast_Buffers.Ressource_Shell.Append ((Kind => Broadcast_Stuff.All_Are_Ok,
                                                          Name => To_Unbounded_String (Name),
                                                          Host_Name => To_Unbounded_String (GNAT.Sockets.Host_Name),
                                                          Process_Name => To_Unbounded_String (Global_Options.Get_Name),
                                                          Error => False,
                                                          Served => False,
                                                          Size => Size,
                                                          Depth => Depth));
            end Provide_Buffer;
         or
            accept Request_Buffer
              (Name : String) do
               Ressource_Name := Null_Unbounded_String;
               Serve_Remotly := False;
               if Buffers.Shell.Memory_Handler.Buffer_Exist (Name) then
                  Ada.Text_IO.Put_Line (Name & " will be served locally");
               else
                  Ada.Text_IO.Put_Line (Name & " should be served remotly");
                  Watchdog := 0;
                  loop
                     Broadcast_Buffers.Ressource_Shell.Ask_For (Name, Broadcast_Stuff.All_Are_Ok);
                     select
                        Broadcast_Buffers.Ressource_Shell.Wait_For_Answer;
                        declare
                           Buffer : constant Broadcast_Stuff.Buffer_Ressource_Type :=
                             Broadcast_Buffers.Ressource_Shell.Get_Ressource (Name);
                        begin
                           Ada.Text_IO.Put_Line (Name & " will be served remotly " & To_String (Buffer.Host_Name));
                           Ressource_Name := To_Unbounded_String (Name);
                           Serve_Remotly := True;
                           exit;
                        end;
                     or
                        delay 1.0;
                        Watchdog := Watchdog + 1;
                        if Watchdog = 5 then
                           raise Broadcast_Stuff.Ressource_Too_Long_To_Answer with "can't get answer for " & Name;
                        end if;
                     end select;
                  end loop;
               end if;
            end Request_Buffer;
            if Serve_Remotly then
               Ada.Text_IO.Put_Line ("start to handle remotly services for " & To_String (Ressource_Name));
               declare
                  Buffer_Name : constant String := To_String (Ressource_Name);
                  Buffer : constant Broadcast_Stuff.Buffer_Ressource_Type :=
                    Broadcast_Buffers.Ressource_Shell.Get_Ressource (Buffer_Name);
                  Local_Ctl : constant String :=
                    Broadcast_Stuff.Get_Process (Pmh_Directory.Directory,
                                                 Broadcast_Stuff.Common_Transport_Layer,
                                                 Global_Options.Get_Name,
                                                 GNAT.Sockets.Host_Name);
               begin
                  Ada.Text_IO.Put_Line ("found local ctl : " & Local_Ctl);
                  Buffers.Shell.Memory_Handler.Create_Buffer (Buffer_Name,
                                                              Container,
                                                              Buffer.Size,
                                                              Buffer.Depth);
                  Container.Message_Handling.Start (Container.Buffer'Access, 1.0);
                  Ada.Text_IO.Put_Line ("local buffer created : " & Buffer_Name);
                  Dcod_Ctl_Service.Client.Request_Client ("TL://" & Buffer_Name,
                                                         Buffer_Name,
                                                         "TL://" & Buffer_Name,
                                                         Local_Ctl);
                  Ada.Text_IO.Put_Line ("local input declaration done");
               exception
                  when E : others =>
                     Ada.Text_IO.Put_Line ("exception in serve remotly : " &
                                             Ada.Exceptions.Exception_Name (E));
                     Ada.Text_IO.Put_Line ("                           : " &
                                             Ada.Exceptions.Exception_Message (E));
               end;
            end if;
         or
            accept Reset_Buffer (Name : String) do
               Buffers.Shell.Memory_Handler.Reset_Buffer (Name);
            end Reset_Buffer;
         or
            accept Delete_Buffer (Name : String) do
               Ada.Text_IO.Put_Line ("trying to delete " & Name);
               declare
                  Ressources : constant Broadcast_Stuff.Buffer_Ressource.Ressource_Vectors.Vector :=
                    Broadcast_Buffers.Ressource_Shell.Get_Ressources;
               begin
                  for Buffer of Ressources loop
                     Ada.Text_IO.Put_Line ("available buffer " & Buffer.Get_Name);
                  end loop;
               end;
               declare
                  Buffer : constant Broadcast_Stuff.Buffer_Ressource_Type :=
                    Broadcast_Buffers.Ressource_Shell.Get_Ressource (Name);
               begin
                  Ada.Text_IO.Put_Line ("buffer description : " & Name);
                  Ada.Text_IO.Put_Line ("                     " & To_String (Buffer.Host_Name));
                  Ada.Text_IO.Put_Line ("                     " & Buffer.Served'Img);
                  if Buffer.Host_Name = GNAT.Sockets.Host_Name then
                     Ada.Text_IO.Put_Line (GNAT.Sockets.Host_Name & " - start to delete locally for " & Name);
                  else
                     Ada.Text_IO.Put_Line (To_String (Buffer.Host_Name) &
                                             " - start to delete remotly for " &
                                             Name);
                     declare
                        Local_Ctl : constant String :=
                          Broadcast_Stuff.Get_Process (Pmh_Directory.Directory,
                                                       Broadcast_Stuff.Common_Transport_Layer,
                                                       Global_Options.Get_Name,
                                                       GNAT.Sockets.Host_Name);
                     begin
                        Dcod_Ctl_Service.Client.Release_Client ("TL://" & Name,
                                                              Local_Ctl);
                        Ada.Text_IO.Put_Line ("local input delete done");
                     end;
                  end if;
                  if Buffer.Served and Buffer.Host_Name = GNAT.Sockets.Host_Name then
                     Ada.Text_IO.Put_Line (GNAT.Sockets.Host_Name & " - start to delete output remotly for " & Name);
                     declare
                        Local_Ctl : constant String :=
                          Broadcast_Stuff.Get_Process (Pmh_Directory.Directory,
                                                       Broadcast_Stuff.Common_Transport_Layer,
                                                       Global_Options.Get_Name,
                                                       GNAT.Sockets.Host_Name);
                     begin
                        Ada.Text_IO.Put_Line ("CTL local output delete call");
                        Dcod_Ctl_Service.Client.Release_Server ("TL://" & Name,
                                                               Local_Ctl);
                        Ada.Text_IO.Put_Line ("CTL local output delete done");
                     end;
                  else
                     Ada.Text_IO.Put_Line ("buffer not served " & Name);
                  end if;
               end;
               if Buffers.Shell.Memory_Handler.Buffer_Exist (Name) then
                  Buffers.Shell.Memory_Handler.Delete_Buffer (Name);
                  Broadcast_Buffers.Ressource_Shell.Remove (Name);
               end if;
            exception
               when Broadcast_Stuff.Buffer_Ressource.Ressource_Not_Found =>
                  Ada.Text_IO.Put_Line (Name & " ressource not found on delete, doing nothing");
               when E : others =>
                  Ada.Text_IO.Put_Line ("Exception in delete buffer : " & Ada.Exceptions.Exception_Name (E));
                  Ada.Text_IO.Put_Line ("                             " & Ada.Exceptions.Exception_Message (E));
                  raise;
            end Delete_Buffer;
         or
            accept Set_Policy (Name : String;
                               Policy : Common_Types.Policy_Type;
                               Value : Boolean) do
               Buffers.Shell.Memory_Handler.Set_Policy (Name,
                                                        Policy,
                                                        Value);
            end Set_Policy;
         or
            accept Set_Mode (Name : String;
                             Mode : Common_Types.Mode_Type;
                             Value : Boolean) do
               Buffers.Shell.Memory_Handler.Set_Mode (Name,
                                                      Mode,
                                                      Value);
            end Set_Mode;
         or
            accept Trig_Event (Event : String) do
               pragma Unreferenced (Event);
               null;
            end Trig_Event;
         or
            terminate;
         end select;
      end loop;
   end Memory_Handler_Task;

   function Get_State return String is
   begin
      return "<error cmd=""get_state"">not implemented</error>";
   end Get_State;

   function Get_Buffer_State (Name : String) return String is
   begin
      return "<error cmd=""get_buffer_state"" name=""" & Name &
        """>not implemented</error>";
   end Get_Buffer_State;

end DCOD.Pmh.Real_Handler;
