with Ada.Containers.Indefinite_Vectors;
with Ada.Strings.Fixed;
with Ada.Strings.Unbounded;
with Ada.Text_IO;

with GNAT.Directory_Operations;
with GNAT.OS_Lib;

with Common_Types.String_Vectors;

package body Dcod_Process_Launcher is

   use Ada.Strings.Unbounded;

   use GNAT.OS_Lib;
   Null_Arg_List : constant Argument_List (1 .. 0) := (others => null);

   function Simplify (Arguments : Argument_List) return Argument_List;
   function Simplify (Arguments : Argument_List) return Argument_List is
      New_Length : Natural := Arguments'Length;
   begin
      for I in Arguments'Range loop
         if Arguments (I).all = "" then
            New_Length := New_Length - 1;
         end if;
      end loop;
      if New_Length = 0 then
         return Null_Arg_List;
      end if;
      declare
         New_Arguments : Argument_List (1 .. New_Length);
         Arg_Index : Positive := 1;
      begin
         for I in Arguments'Range loop
            if not (Arguments (I).all = "") then
               New_Arguments (Arg_Index) := Arguments (I);
               Arg_Index := Arg_Index + 1;
            end if;
         end loop;
         return New_Arguments;
      end;
   end Simplify;

   type Launched_Process_Container_Type (Parameters_Length : Natural) is record
      Process_Name : Unbounded_String;
      Parameters : Common_Types.Strings.String_Couple_Array
        (1 .. Parameters_Length);
      Process_Id : GNAT.OS_Lib.Process_Id;
   end record;
   package Launched_Process_Container_Vectors is new
     Ada.Containers.Indefinite_Vectors (Positive,
                                        Launched_Process_Container_Type,
                                        "=");

   protected Launcher_Protected is
      procedure Real_Append_Directory (Name : String);
      procedure Real_Launch
        (Process_Name : String;
         Parameters : Common_Types.Strings.String_Couple_Array);
      function Real_Get_Executable_List
        (Directory_Name : String) return Common_Types.Strings.String_Array;
      function Real_Get_Running_List
        return Common_Types.Strings.String_Couple_Array;
      function Real_Get_Properties
        (Process_Name : String)
        return Common_Types.Strings.String_Couple_Array;
      procedure Real_Remove
        (Process_Name : String;
         Parameters : Common_Types.Strings.String_Couple_Array;
         Kill : Boolean);
      procedure Real_Remove_By_Pid (Process_Name : String;
                                    Pid : Integer;
                                    Kill : Boolean);
      procedure Real_Set_Debug (Debug_Arg : Boolean);
   private
      Debug : Boolean := False;
      Directories : Common_Types.String_Vectors.Vector;
      Launched_Processes : Launched_Process_Container_Vectors.Vector;
   end Launcher_Protected;

   ----------------------
   -- Append_Directory --
   ----------------------

   procedure Append_Directory (Directory_Name : String) is
   begin
      Launcher_Protected.Real_Append_Directory (Directory_Name);
   end Append_Directory;

   procedure Set_Debug (Debug : Boolean) is
   begin
      Launcher_Protected.Real_Set_Debug (Debug);
   end Set_Debug;

   ------------
   -- Launch --
   ------------

   procedure Launch
     (Process_Name : String;
      Parameters : Common_Types.Strings.String_Couple_Array)
   is
   begin
      Launcher_Protected.Real_Launch (Process_Name, Parameters);
   end Launch;

   -------------------------
   -- Get_Executable_List --
   -------------------------

   function Get_Executable_List
     (Directory_Name : String)
      return Common_Types.Strings.String_Array
   is
   begin
      return Launcher_Protected.Real_Get_Executable_List (Directory_Name);
   end Get_Executable_List;

   ----------------------
   -- Get_Running_List --
   ----------------------

   function Get_Running_List
      return Common_Types.Strings.String_Couple_Array
   is
   begin
      return Launcher_Protected.Real_Get_Running_List;
   end Get_Running_List;

   --------------------
   -- Get_Properties --
   --------------------

   function Get_Properties
     (Process_Name : String)
      return Common_Types.Strings.String_Couple_Array
   is
   begin
      return Launcher_Protected.Real_Get_Properties (Process_Name);
   end Get_Properties;

   ------------
   -- Remove --
   ------------

   procedure Remove
     (Process_Name : String;
      Parameters : Common_Types.Strings.String_Couple_Array;
      Kill : Boolean)
   is
   begin
      Launcher_Protected.Real_Remove (Process_Name, Parameters, Kill);
   end Remove;

   procedure Remove_By_Pid
     (Process_Name : String;
      Pid : Integer;
      Kill : Boolean)
   is
   begin
      Launcher_Protected.Real_Remove_By_Pid (Process_Name, Pid, Kill);
   end Remove_By_Pid;

   protected body Launcher_Protected is

      procedure Put_Line (Message : String);

      procedure Put_Line (Message : String) is
      begin
         if not Debug then
            return;
         end if;
         Ada.Text_IO.Put_Line (Message);
      end Put_Line;

      procedure Real_Append_Directory (Name : String) is
      begin
         if Is_Directory (Name) then
            Directories.Append (Name);
         else
            raise API_Misused with Name & " isn't a directory";
         end if;
      end Real_Append_Directory;
      procedure Real_Launch
        (Process_Name : String;
         Parameters : Common_Types.Strings.String_Couple_Array) is
         Launch_Not_Done : Boolean := True;
         Slash_Position : constant Natural :=
           Ada.Strings.Fixed.Index (Process_Name, "/");
         Dot_Position : constant Natural :=
           Ada.Strings.Fixed.Index (Process_Name, ".");
         use Common_Types.Strings;
         Pid : Process_Id;
      begin
         if Slash_Position /= 0 then
            raise Bad_Process_Name with
              "no slash allowed in process name : " & Process_Name;
         end if;
         if Dot_Position /= 0 then
            raise Bad_Process_Name with
              "no dot allowed in process name : " & Process_Name;
         end if;
         for Directory of Directories loop
            Put_Line ("directory");
            if Is_Executable_File (Directory & "/" & Process_Name) then
               Launch_Not_Done := False;
               Put_Line (Directory & "/" & Process_Name);
               if Parameters = Null_String_Couple_Array then
                  Put_Line ("no arg");
                  Pid := Non_Blocking_Spawn (Directory & "/" & Process_Name,
                                             Null_Arg_List);
                  Launched_Processes.Append
                       ((Parameters_Length => 0,
                         Process_Name => To_Unbounded_String (Process_Name),
                         Parameters => <>,
                         Process_Id => Pid));
               else
                  Put_Line ("args");
                  declare
                     Arg_List : Argument_List (1 .. Parameters'Length * 2);
                  begin
                     for I in Parameters'Range loop
                        Arg_List (2 * I - 1) :=
                          new String'(To_String (Parameters (I).Name));
                        Put_Line ("name " & Arg_List (2 * I - 1).all);
                        Arg_List (2 * I) :=
                          new String'(To_String (Parameters (I).Value));
                        Put_Line ("value " & Arg_List (2 * I).all);
                     end loop;
                     Put_Line ("end of args");
                     Pid := Non_Blocking_Spawn (Directory & "/" & Process_Name,
                                                Simplify (Arg_List));
                     Launched_Processes.Append
                       ((Parameters_Length => Parameters'Length,
                         Process_Name => To_Unbounded_String (Process_Name),
                         Parameters => Parameters,
                         Process_Id => Pid));
                     for I in Arg_List'Range loop
                        Free (Arg_List (I));
                     end loop;
                  end;
               end if;
            end if;
         end loop;
         if Launch_Not_Done then
            raise Executable_Not_Found with
              Process_Name & " not found in all directories";
         end if;
      end Real_Launch;
      function Real_Get_Executable_List
        (Directory_Name : String) return Common_Types.Strings.String_Array is
         Executables : Common_Types.String_Vectors.Vector;
         procedure Search_One_Directory (Name : String);
         procedure Search_One_Directory (Name : String) is
            use GNAT.Directory_Operations;
            Directory : Dir_Type;
            Executable_Name : String (1 .. 128);
            Last : Natural;
         begin
            Open (Directory, Name);
            loop
               Read (Directory, Executable_Name, Last);
               exit when Last = 0;
               declare
                  Full_Name : constant String :=
                    Name & "/" & Executable_Name (1 .. Last);
               begin
                  if Is_Executable_File (Full_Name) then
                     if not Is_Directory (Full_Name) then
                        Executables.Append (Executable_Name (1 .. Last));
                     end if;
                  end if;
               end;
            end loop;
            Close (Directory);
         end Search_One_Directory;
         Length : Natural;
      begin
         if Directory_Name = "#ALL_DIRECTORIES#" then
            for Directory of Directories loop
               Search_One_Directory (Directory);
            end loop;
         else
            if not Is_Directory (Directory_Name) then
               raise API_Misused with Directory_Name & " isn't a directory";
            end if;
            for Directory of Directories loop
               if Directory = Directory_Name then
                  Search_One_Directory (Directory_Name);
               end if;
            end loop;
         end if;
         Length := Natural (Executables.Length);
         if Length = 0 then
            return Common_Types.Strings.Null_String_Array;
         else
            declare
               Executables_Array : Common_Types.Strings.String_Array
                 (1 .. Length);
               Exe_Index : Positive := 1;
            begin
               for Executable of Executables loop
                  Executables_Array (Exe_Index) :=
                    To_Unbounded_String (Executable);
                  Exe_Index := Exe_Index + 1;
               end loop;
               Executables.Clear;
               return Executables_Array;
            end;
         end if;
      end Real_Get_Executable_List;
      function Real_Get_Running_List
        return Common_Types.Strings.String_Couple_Array is
         Length : constant Natural := Natural (Launched_Processes.Length);
         use Common_Types.Strings;
         Running_Executables : String_Couple_Array (1 .. Length);
         Running_Index : Positive := 1;
      begin
         for Executable of Launched_Processes loop
            Running_Executables (Running_Index) :=
              (Name => Executable.Process_Name,
               Value => To_Unbounded_String (Pid_To_Integer
                                               (Executable.Process_Id)'Img));
            Running_Index := Running_Index + 1;
         end loop;
         if Length = 0 then
            return Null_String_Couple_Array;
         else
            return Running_Executables;
         end if;
      end Real_Get_Running_List;
      function Real_Get_Properties
        (Process_Name : String)
        return Common_Types.Strings.String_Couple_Array is
         pragma Unreferenced (Process_Name);
      begin
         return Common_Types.Strings.Null_String_Couple_Array;
      end Real_Get_Properties;
      procedure Real_Remove
        (Process_Name : String;
         Parameters : Common_Types.Strings.String_Couple_Array;
         Kill : Boolean) is
         Proc_Index : Natural := 0;
         Current_Index : Positive := 1;
         use type Common_Types.Strings.String_Couple_Array;
         Wait_Success : Boolean;
      begin
         for Launched_Process of Launched_Processes loop
            if Launched_Process.Process_Name = Process_Name and Launched_Process.Parameters = Parameters then
               Proc_Index := Current_Index;
               if Kill then
                  declare
                     Pid_Integer : constant Integer :=
                       Pid_To_Integer (Launched_Process.Process_Id);
                     Arg_List : Argument_List (1 .. 2) :=
                       (new String'("-9"), new String'(Pid_Integer'Img));
                     Success : Boolean;
                  begin
                     Spawn ("/bin/kill", Arg_List, Success);
                     for I in Arg_List'Range loop
                        Free (Arg_List (I));
                     end loop;
                  end;
               end if;
               Wait_Process (Launched_Process.Process_Id, Wait_Success);
            end if;
            Current_Index := Current_Index + 1;
         end loop;
         if Proc_Index /= 0 then
            Launched_Processes.Delete (Proc_Index);
         end if;
      end Real_Remove;
      procedure Real_Remove_By_Pid (Process_Name : String;
                                    Pid : Integer;
                                    Kill : Boolean) is
         Proc_Index : Natural := 0;
         Current_Index : Positive := 1;
         use type Common_Types.Strings.String_Couple_Array;
         Wait_Success : Boolean;
         Out_Process_Id : Process_Id;
      begin
         Ada.Text_IO.Put_Line ("remove by pid : " & Process_Name & Pid'Img);
         for Launched_Process of Launched_Processes loop
            if Launched_Process.Process_Name = Process_Name and Pid_To_Integer (Launched_Process.Process_Id) = Pid then
               Proc_Index := Current_Index;
               Ada.Text_IO.Put_Line ("remove by pid, process found");
               if Kill then
                  declare
                     Pid_Integer : constant Integer :=
                       Pid_To_Integer (Launched_Process.Process_Id);
                     Arg_List : Argument_List (1 .. 2) :=
                       (new String'("-9"), new String'(Pid_Integer'Img));
                     Success : Boolean;
                  begin
                     Spawn ("kill", Arg_List, Success);
                     Ada.Text_IO.Put_Line ("remove by pid," &
                                             Pid_Integer'Img &
                                             " kill return " &
                                             Success'Img);
                     for I in Arg_List'Range loop
                        Free (Arg_List (I));
                     end loop;
                  end;
               end if;
               Ada.Text_IO.Put_Line ("remove by pid, start waiting");
               Wait_Process (Out_Process_Id, Wait_Success);
               Ada.Text_IO.Put_Line ("remove by pid" &
                                       Pid_To_Integer (Out_Process_Id)'Img &
                                       ", waiting ended " &
                                       Wait_Success'Img);
            end if;
            Current_Index := Current_Index + 1;
         end loop;
         if Proc_Index /= 0 then
            Launched_Processes.Delete (Proc_Index);
         end if;
      end Real_Remove_By_Pid;

      procedure Real_Set_Debug (Debug_Arg : Boolean) is
      begin
         Debug := Debug_Arg;
      end Real_Set_Debug;

   end Launcher_Protected;

end Dcod_Process_Launcher;
